# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from ..core import ScalarField, VectorField, \
    TemporalVectorFields, TemporalScalarFields,\
    Field, Profile, TemporalFields
from ..utils import make_unit
from ..utils.types import ARRAYTYPES, NUMBERTYPES, STRINGTYPES
import numpy as np
try:
    import modred
    MODRED = True
except ImportError:
    MODRED = False
import matplotlib.pyplot as plt
from .. import plotlib as pplt
import scipy.integrate as spint


class ModalFields(Field):
    """
    Class representing the result of a modal decomposition.
    """

    def __init__(self, decomp_type, mean_field, modes, modes_numbers,
                 temporal_evolutions,
                 eigvals=None, eigvects=None, ritz_vals=None, mode_norms=None,
                 growth_rate=None, pulsation=None):
        """
        Constructor
        """
        # Check if modred available
        if not MODRED:
            raise Exception("This feature need 'modred' to be installed")
        Field.__init__(self)
        # check parameters
        if decomp_type not in ['pod', 'dmd', 'bpod']:
            raise ValueError()
        self.field_class = mean_field.__class__
        if not isinstance(modes, ARRAYTYPES):
            raise TypeError()
        modes = np.array(modes)
        if not isinstance(modes[0], self.field_class):
            raise TypeError()
        if not isinstance(modes_numbers, ARRAYTYPES):
            raise TypeError()
        modes_numbers = np.array(modes_numbers)
        if not modes_numbers.dtype == int:
            raise TypeError()
        if not len(modes_numbers) == len(modes):
            raise ValueError()
        if not isinstance(modes_numbers, ARRAYTYPES):
            raise TypeError()
        modes_numbers = np.array(modes_numbers)
        if not isinstance(temporal_evolutions[0], Profile):
            raise TypeError()
        if not len(temporal_evolutions) == len(modes):
            raise ValueError()
        if eigvals is not None:
            if not isinstance(eigvals, Profile):
                raise TypeError()
            if not len(eigvals) == len(modes):
                raise ValueError()
        if eigvects is not None:
            if not isinstance(eigvects, ARRAYTYPES):
                raise TypeError()
            eigvects = np.array(eigvects)
            if not eigvects.shape == (len(temporal_evolutions[0].x),
                                      len(modes)):
                raise ValueError()
        if ritz_vals is not None:
            if not isinstance(ritz_vals, Profile):
                raise TypeError()
            if not len(ritz_vals) == len(modes):
                raise ValueError()
        if mode_norms is not None:
            if not isinstance(mode_norms, Profile):
                raise TypeError()
            if not len(mode_norms) == len(modes):
                raise ValueError()
        if growth_rate is not None:
            if not isinstance(growth_rate, Profile):
                raise TypeError()
            if not len(growth_rate) == len(modes):
                raise ValueError()
        if pulsation is not None:
            if not isinstance(pulsation, Profile):
                raise TypeError()
            if not len(pulsation) == len(modes):
                raise ValueError()
        # storing
        self.decomp_type = decomp_type
        self.mean_field = mean_field
        self.axe_x = mean_field.axe_x
        self.axe_y = mean_field.axe_y
        self.unit_x = mean_field.unit_x
        self.unit_y = mean_field.unit_y
        self.unit_values = mean_field.unit_values
        self.modes = modes
        self.modes_nmb = modes_numbers
        self.temp_evo = temporal_evolutions
        self.times = temporal_evolutions[0].x
        self.unit_times = temporal_evolutions[0].unit_x
        if eigvals is not None:
            self.eigvals = eigvals
        if eigvects is not None:
            self.eigvects = eigvects
        if ritz_vals is not None:
            self.ritz_vals = ritz_vals
        if mode_norms is not None:
            self.mode_norms = mode_norms
        if growth_rate is not None:
            self.growth_rate = growth_rate
        if pulsation is not None:
            self.pulsation = pulsation
        self.__temp_coh = None
        self.__spat_coh = None

    @property
    def modes_as_tf(self):
        if self.field_class == VectorField:
            tmp_tf = TemporalVectorFields()
            for i in np.arange(len(self.modes)):
                tmp_tf.add_field(self.modes[i], time=i + 1,
                                 unit_times="")
            return tmp_tf
        elif self.field_class == ScalarField:
            tmp_tf = TemporalScalarFields()
            for i in np.arange(len(self.modes)):
                tmp_tf.add_field(self.modes[i], time=i + 1,
                                 unit_times="")
            return tmp_tf

    def crop_modal_base(self, modes_to_keep=None, modes_to_remove=None,
                        inplace=True):
        """
        Remove some modes from the modal base.

        Parameters
        ----------
        modes_to_keep : array of integers or integer
            If an integer, the first N modes are kept, if an array of indices,
            all the associated modes are conserved.
        modes_to_remove :  array of integers or integer
            If an integer, the last N modes are removed,
            if an array of indices, all the associated modes are removed.
        """
        # check
        if modes_to_keep is None and modes_to_remove is None:
            raise ValueError()
        if modes_to_keep is not None:
            if isinstance(modes_to_keep, int):
                modes_to_keep = np.arange(modes_to_keep)
            if not isinstance(modes_to_keep, ARRAYTYPES):
                raise TypeError()
            modes_to_keep = np.array(modes_to_keep, dtype=int)
        else:
            if isinstance(modes_to_remove, int):
                modes_to_remove = np.arrange(len(self.modes) - modes_to_remove,
                                             len(self.modes))
            if not isinstance(modes_to_remove, ARRAYTYPES):
                raise TypeError()
            modes_to_remove = np.array(modes_to_remove, dtype=int)
            modes_to_keep = np.arrange(len(self.modes))
            modes_to_keep = np.delete(modes_to_keep, modes_to_remove)
        # get data
        if inplace:
            tmp_pod = self
        else:
            tmp_pod = self.copy()
        # prepare new containers
        new_modes = []
        new_temp_evo = []
        if self.decomp_type == 'pod':
            new_eigvals = Profile(unit_x=self.eigvals.unit_x,
                                  unit_y=self.eigvals.unit_y)
            new_eigvects = []
        elif self.decomp_type == "dmd":
            new_growth_rate = Profile(unit_x=self.eigvals.unit_x,
                                      unit_y=self.eigvals.unit_y)
            new_mode_norms = Profile(unit_x=self.eigvals.unit_x,
                                     unit_y=self.eigvals.unit_y)
            new_pulsation = Profile(unit_x=self.eigvals.unit_x,
                                    unit_y=self.eigvals.unit_y)
            new_ritz_vals = Profile(unit_x=self.eigvals.unit_x,
                                    unit_y=self.eigvals.unit_y)
        else:
            raise Exception('Not implemented for {}'.format(self.decomp_type))
        # loop on wanted modes
        for ind in modes_to_keep:
            new_modes.append(tmp_pod.modes[ind])
            new_temp_evo.append(tmp_pod.temp_evo[ind])
            if self.decomp_type == 'pod':
                new_eigvals.add_point(tmp_pod.eigvals.x[ind],
                                      tmp_pod.eigvals.y[ind])
                tmp_vect = []
                for ind2 in modes_to_keep:
                    tmp_vect.append(tmp_pod.eigvects[ind, ind2])
                new_eigvects.append(tmp_vect)
            if self.decomp_type == "dmd":
                new_growth_rate.add_point(tmp_pod.growth_rate.x[ind],
                                          tmp_pod.growth_rate.y[ind])
                new_mode_norms.add_point(tmp_pod.mode_norms.x[ind],
                                         tmp_pod.mode_norms.y[ind])
                new_pulsation.add_point(tmp_pod.pulsation.x[ind],
                                        tmp_pod.pulsation.y[ind])
                new_ritz_vals.add_point(tmp_pod.ritz_vals.x[ind],
                                        tmp_pod.ritz_vals.y[ind])
        # store
        tmp_pod.modes_nmb = tmp_pod.modes_nmb[modes_to_keep]
        tmp_pod.modes = new_modes
        tmp_pod.temp_evo = new_temp_evo
        if self.decomp_type == 'pod':
            tmp_pod.eigvals = new_eigvals
            tmp_pod.eigvects = np.array(new_eigvects, dtype=float)
        if self.decomp_type == "dmd":
            tmp_pod.growth_rate = new_growth_rate
            tmp_pod.mode_norms = new_mode_norms
            tmp_pod.pulsation = new_pulsation
            tmp_pod.ritz_vals = new_ritz_vals
        tmp_pod._ModalFields__spat_coh = None
        tmp_pod._ModalFields__temp_coh = None
        # return
        if not inplace:
            return tmp_pod

    def crop(self, intervx=None, intervy=None, intervt=None, ind=False,
             inplace=False):
        """
        Crop the POD modes and/or temporal evolutions, according to the given
        intervals.

        Parameters
        ----------
        intervx : 2x1 array of numbers
            Interval of x to keep.
        intervy : 2x1 array of numbers
            Interval of y to keep.
        intervt : 2x1 array of numbers
            Interval of time to keep.
        ind : boolean, optional
            If 'True', intervals are understood as indices along axis.
            If 'False' (default), intervals are understood in axis units.
        inplace : boolean, optional
            If 'True', the field is croped in place.
        """
        # check not necessary because done in 'VectorField.crop()'
        # get data
        if inplace:
            tmp_mf = self
        else:
            tmp_mf = self.copy()
        # temporal crop (first for efficiency)
        if intervt is not None:
            times = self.times
            # get time indices
            if not ind:
                intervt = np.array(intervt, dtype=float)
                if intervt[0] <= times[0]:
                    ind_1 = 0
                elif intervt[0] >= times[-1]:
                    raise ValueError()
                else:
                    ind_1 = np.where(times > intervt[0])[0][0]
                if intervt[1] >= times[-1]:
                    ind_2 = len(times)
                elif intervt[1] <= times[0]:
                    raise ValueError()
                else:
                    ind_2 = np.where(times > intervt[1])[0][0]
                intervt = [ind_1, ind_2]
            intervt = np.array(intervt, dtype=int)
            # crop
            tmp_mf.times = tmp_mf.times[intervt[0]:intervt[1]]
            tmp_mf._ModalFields__temp_coh = None
            for prof in tmp_mf.temp_evo:
                prof.crop(intervx=intervt, ind=True, inplace=True)
        # spatial crop
        if intervx is not None or intervy is not None:
            # crop Field and mean field
            super(ModalFields, tmp_mf).crop(intervx=intervx, intervy=intervy,
                                            full_output=False, ind=ind,
                                            inplace=True)
            tmp_mf.mean_field.crop(intervx=intervx, intervy=intervy, ind=ind,
                                   inplace=True)
            # loop on modes
            for mode in tmp_mf.modes:
                mode.crop(intervx=intervx, intervy=intervy,
                          ind=ind, inplace=True)
        # return
        if not inplace:
            return tmp_mf

    def scale(self, scalex=1., scaley=1., scalet=1., scalev=1.,
              inplace=False):
        """
        """
        #
        if inplace:
            tmp_mf = self
        else:
            tmp_mf = self.copy()
        # scale mean field
        tmp_mf.mean_field.scale(scalex=scalex, scaley=scaley, scalev=scalev,
                                inplace=True)
        # scale axe and unities
        tmp_mf.axe_x = tmp_mf.mean_field.axe_x
        tmp_mf.axe_y = tmp_mf.mean_field.axe_y
        tmp_mf.unit_x = tmp_mf.mean_field.unit_x
        tmp_mf.unit_y = tmp_mf.mean_field.unit_y
        tmp_mf.unit_values = tmp_mf.mean_field.unit_values
        # scale temporal evolutions
        for i in range(len(tmp_mf.temp_evo)):
            tmp_mf.temp_evo[i].scale(scalex=scalet, scaley=scalev,
                                     inplace=True)
        # scale times and unit
        tmp_mf.times = tmp_mf.temp_evo[0].x
        tmp_mf.unit_times = tmp_mf.temp_evo[0].unit_x
        # scale modes
        for i in range(len(tmp_mf.modes)):
            tmp_mf.modes[i].scale(scalex=scalex, scaley=scaley,
                                  inplace=True)
        # return
        if not inplace:
            return tmp_mf

    def smooth_temporal_evolutions(self, tos='uniform', size=None,
                                   inplace=True):
        """
        Smooth the temporal evolutions (to do before reconstructing)

        Parameters
        ----------
        tos : string, optional
            Type of smoothing, can be ‘uniform’ (default) or ‘gaussian’
            (See ndimage module documentation for more details)
        size : number, optional
            Size of the smoothing
            (is radius for ‘uniform’ and sigma for ‘gaussian’).
            Default is 3 for ‘uniform’ and 1 for ‘gaussian’.
        """
        if inplace:
            for n in np.arange(len(self.modes)):
                self.temp_evo[n].smooth(tos=tos, size=size, inplace=True)
        else:
            tmp_MF = self.copy()
            for n in np.arange(len(tmp_MF.modes)):
                tmp_MF.temp_evo[n].smooth(tos=tos, size=size, inplace=True)
            return tmp_MF

    def smooth_spatial_evolutions(self, tos='uniform', size=None, inplace=True):
        """
        Smooth the spatial evolutions (to do before reconstructing)

        Parameters
        ----------
        tos : string, optional
            Type of smoothing, can be ‘uniform’ (default) or ‘gaussian’
            (See ndimage module documentation for more details)
        size : number, optional
            Size of the smoothing
            (is radius for ‘uniform’ and sigma for ‘gaussian’).
            Default is 3 for ‘uniform’ and 1 for ‘gaussian’.
        """
        if inplace:
            for n in np.arange(len(self.modes)):
                self.modes[n].smooth(tos=tos, size=size, inplace=True)
        else:
            tmp_MF = self.copy()
            for n in np.arange(len(tmp_MF.modes)):
                tmp_MF.mdes[n].smooth(tos=tos, size=size, inplace=True)
            return tmp_MF

    def augment_temporal_resolution(self, fact=2, interp='linear',
                                    inplace=True, verbose=False):
        """
        Augment the temporal resolution (to augmente the temporal resolution
        of the reconstructed fields).

        Parameters
        ----------
        fact : integer
            Resolution augmentation needed (default is '2', for a result
            profile with twice more points)
        interp : string in ['linear', 'nearest', 'slinear', 'quadratic',
                            'cubic']
            Specifies the kind of interpolation as a string
            (Default is 'linear'). slinear', 'quadratic' and 'cubic' refer
            to a spline interpolation of first, second or third order.
        inplace : bool
            .
        verbose : bool
            .
        """
        # check parameters
        if not isinstance(fact, int):
            if int(fact) - fact == 0:
                fact = int(fact)
            else:
                raise TypeError()
        if fact <= 0:
            raise TypeError()
        if not isinstance(interp, STRINGTYPES):
            raise TypeError()
        if interp not in ['linear', 'nearest', 'slinear', 'quadratic',
                          'cubic']:
            raise ValueError()
        if not isinstance(inplace, bool):
            raise TypeError()
        # get data
        if inplace:
            tmp_pod = self
        else:
            tmp_pod = self.copy()
        # loop on the temporal evolutions
        for prof in tmp_pod.temp_evo:
            prof.augment_resolution(fact=fact, interp=interp, inplace=True)
        # change other little things
        tmp_pod.times = tmp_pod.temp_evo[0].x
        tmp_pod._ModalFields__spat_coh = None
        tmp_pod._ModalFields__temp_coh = None
        # return
        if not inplace:
            return tmp_pod

    def reconstruct(self, wanted_modes='all', times=None):
        """
        Recontruct fields resolved in time from modes.

        Parameters
        ----------
        wanted_modes : string or number or array of numbers, optional
            wanted modes for reconstruction :
            If 'all' (default), all modes are used
            If an array of integers, the wanted modes are used
            If an integer, the wanted first modes are used.
        times : tuple of numbers
            If specified, reconstruction is computed on the wanted times,
            else, times used for decomposition are used.
        Returns
        -------
        TF : TemporalFields (TemporalScalarFields or TemporalVectorFields)
            Reconstructed fields.
        """
        # check parameters
        if isinstance(wanted_modes, STRINGTYPES):
            if wanted_modes == 'all':
                wanted_modes = np.arange(len(self.modes))
        elif isinstance(wanted_modes, NUMBERTYPES):
            if self.decomp_type in ['pod', 'bpod']:
                wanted_modes = np.arange(wanted_modes)
            elif self.decomp_type == 'dmd':
                wanted_modes = (np.argsort(np.abs(self.growth_rate.y))
                                [0:wanted_modes])
            else:
                raise ValueError()
        elif isinstance(wanted_modes, ARRAYTYPES):
            wanted_modes = np.array(wanted_modes)
            if not isinstance(wanted_modes[0], NUMBERTYPES):
                raise TypeError()
            if wanted_modes.max() > len(self.modes):
                raise ValueError()
        else:
            raise TypeError()
        if times is None:
            times = self.times
        if not isinstance(times, ARRAYTYPES):
            raise TypeError()
        times = np.array(times)
        if times.ndim != 1:
            raise ValueError()
        if self.decomp_type in ['pod', 'bpod']:
            if (times.max() > self.times.max() or
                    times.min() < self.times.min()):
                raise ValueError()
        # getting datas
        ind_times = np.arange(len(times))
        super_mask = self.modes[0].mask
        # reconstruction temporal evolution if needed
        if self.decomp_type in ['pod', 'bpod']:
            temp_evo = self.temp_evo
        elif self.decomp_type == 'dmd' and np.all(self.times == times):
            temp_evo = self.temp_evo
        elif self.decomp_type == 'dmd' and not np.all(self.times == times):
            temp_evo = []
            delta_t1 = self.times[1] - self.times[0]
            ks = times/delta_t1
            for n in np.arange(len(self.modes)):
                tmp_prof = [self.ritz_vals.y[n]**(k - 1) for k in ks]
                tmp_prof = Profile(times, tmp_prof, mask=False,
                                   unit_x=self.unit_times,
                                   unit_y=self.unit_values)
                temp_evo.append(tmp_prof)
        # TSF
        if self.field_class == ScalarField:
            # mean field
            tmp_tf = np.array([self.mean_field.values]*len(times))
            # loop on the modes
            for n in wanted_modes:
                tmp_mode = self.modes[n].values
                tmp_prof = temp_evo[n]
                for t in ind_times:
                    coef = tmp_prof.get_interpolated_values(x=times[t])[0]
                    tmp_tf[t] += np.real(tmp_mode*coef)
            # returning
            TF = TemporalScalarFields()
            for t in ind_times:
                tmp_sf = ScalarField()
                tmp_sf.import_from_arrays(self.axe_x, self.axe_y, tmp_tf[t],
                                          mask=super_mask,
                                          unit_x=self.unit_x,
                                          unit_y=self.unit_y,
                                          unit_values=self.unit_values)
                TF.add_field(tmp_sf, time=times[t], unit_times=self.unit_times)
        # TVF
        elif self.field_class == VectorField:
            # mean field
            tmp_tf_x = np.array([self.mean_field.comp_x]*len(times))
            tmp_tf_y = np.array([self.mean_field.comp_y]*len(times))
            # loop on the modes
            for n in wanted_modes:
                tmp_mode_x = self.modes[n].comp_x
                tmp_mode_y = self.modes[n].comp_y
                tmp_prof = temp_evo[n]
                for t in ind_times:
                    coef = tmp_prof.get_interpolated_values(x=times[t])[0]
                    tmp_tf_x[t] += np.real(tmp_mode_x*coef)
                    tmp_tf_y[t] += np.real(tmp_mode_y*coef)
            # returning
            TF = TemporalVectorFields()
            for t in ind_times:
                tmp_vf = VectorField()
                tmp_vf.import_from_arrays(self.axe_x, self.axe_y,
                                          tmp_tf_x[t], tmp_tf_y[t],
                                          mask=super_mask,
                                          unit_x=self.unit_x,
                                          unit_y=self.unit_y,
                                          unit_values=self.unit_values)
                TF.add_field(tmp_vf, time=times[t], unit_times=self.unit_times)
        return TF

    def get_modes_energy(self, cum=False, raw=False):
        """
        Return a profile whith the modes mean energy.

        Parameters
        ----------
        cum : boolean, optional
            If 'False' (default), return the modes energy
            If 'True', return the cumulative modes energy
        raw : bool, optional
            If 'False' (default), a Profile object is returned
            If 'True', an array is returned

        Returns
        -------
        modes_nrj : array or Profile object
            Energy for each modes.

        """
        modes_nrj = np.zeros((len(self.modes),))
        for n in np.arange(len(self.modes)):
            if isinstance(self.modes[0], ScalarField):
                magnitude = 1./2.*np.real(self.modes[n].values)**2
            elif isinstance(self.modes[0], VectorField):
                magnitude = 1./2.*(np.real(self.modes[n].comp_x)**2 +
                                   np.real(self.modes[n].comp_y)**2)
            coef_temp = np.mean(np.real(self.temp_evo[n].y)**2)
            modes_nrj[n] = np.mean(magnitude)*coef_temp
        # cum or not
        if cum:
            modes_nrj = np.cumsum(modes_nrj)
        # returning
        if raw:
            return modes_nrj
        else:
            modes_nrj = Profile(np.arange(len(self.modes)), modes_nrj,
                                mask=False,
                                unit_x="", unit_y=self.modes[0].unit_values**2,
                                name="")
        return modes_nrj

    def get_critical_kappa(self, Nx, Ny=None):
        """
        Return the critical value of kappa.
        A mode with a kappa value superior to the critical value have only
        .3% chance to be a random mode.
        """
        if Ny is None:
            a = 9.57253208
            b = 0.92940704
            c = 0.98559246
            kappa_crit = np.exp(a/Nx**b) - c
            return kappa_crit
        else:
            a, b, c = (22.885989, 1.552608, 1.996703)
            return np.exp(a/Nx**b) + np.exp(a/Ny**b) - c

    def get_temporal_coherence(self, raw=False):
        """
        Return a profile where each value represent the probability for a mode
        to be temporaly non-random (values above 0 have only 5% chance to be
        associated with random modes).
        Can be used to determine the modes to take to filter the turbulence
        (and so perform a tri-decomposition (mean + coherent + turbulent))

        Parameters
        ----------
        raw : bool, optional
            If 'False' (default), a Profile object is returned
            If 'True', an array is returned

        Returns
        -------
        var_spec : array or Profile object
            Probability estimation for each mode of being coherent in time.

        Notes
        -----
        Returned values are, for each modes, the variance of the normalized
        spectrum of the temporal evolution.
        Variance is high when spectrum show predominant frequencies
        (coherent behavior), inversely, variance is low when spectrum is
        nearly uniform (random behavior).
        """
        if self._ModalFields__temp_coh is not None:
            var_spec = self._ModalFields__temp_coh
        else:
            # computing maximal variance
            max_std_spec = np.zeros(len(self.times))
            max_std_spec[0] = 1
            max_std_spec /= np.trapz(max_std_spec)
            max_std = np.std(max_std_spec)
            # computing critical kappa
            k_crit = self.get_critical_kappa(len(self.temp_evo[0].x))
            # getting spectrum variance on modes
            var_spec = np.empty((len(self.modes)))
            for n in np.arange(len(self.modes)):
                prof = self.temp_evo[n]
                spec = prof.get_spectrum(scaling='spectrum')**.5
                spec /= np.trapz(spec.y)
                var_spec[n] = np.std(spec.y)/max_std
                if var_spec[n] < 0:
                    var_spec[n] = 0
            # normalize with critical kappa
            var_spec = (var_spec - k_crit)/(1 - k_crit)
            # storing on object
            self._ModalFields__temp_coh = var_spec
        # returning
        if raw:
            return var_spec
        else:
            prof = Profile(np.arange(len(self.modes)), var_spec, unit_x='',
                           unit_y='', name="")
            return prof

    def get_spatial_coherence(self, raw=False):
        """
        Return a profile where each value represent the probability for a mode
        to be spatialy non-random (values from 0 to 1).
        Can be used to determine the modes to take to filter the turbulence
        (and so perform a tri-decomposition (mean + coherent + turbulent))

        Parameters
        ----------
        raw : bool, optional
            If 'False' (default), a Profile object is returned
            If 'True', an array is returned

        Returns
        -------
        var_spec : array or Profile object
            Probability estimation for each mode of being coherent in space.

        Notes
        -----
        Returned values are, for each modes, the variance of the
        normalized two-dimensional spectrum of Vx and Vy.
        Variance is high when spectrum show predominant frequencies
        (coherent behavior), inversely, variance is low when spectrum is
        nearly uniform (random behavior).
        """
        if self._ModalFields__spat_coh is not None:
            var_spec = self._ModalFields__spat_coh
        else:
            # data
            shape = self.modes[0].shape
            axe_x = self.modes[0].axe_x
            axe_y = self.modes[0].axe_y
            center_x = np.int(len(axe_x)/2.)
            center_y = np.int(len(axe_y)/2.)

            def get_field_spec_std(data):
                data -= np.mean(data)
                spec = np.abs(np.real(np.fft.rfft2(data)))
                spec = np.fft.fftshift(spec)
                spec /= spint.simps(spint.simps(spec))
                return np.std(spec)
            # computing minimal variance
            min_std = 0.
            for i in np.arange(20):
                min_std_field = np.random.rand(*shape)*2. - 1.
                min_std += get_field_spec_std(min_std_field)
            min_std /= 20.
            # computing maximal variance
            max_std_spec = np.zeros(shape)
            max_std_spec[center_x:center_x + 2, center_y:center_y + 2] = 1.
            max_std_spec /= spint.simps(spint.simps(max_std_spec))
            max_std = np.std(max_std_spec)
#            # computing critical kappa
#            k_crit = self.get_critical_kappa(len(self.axe_x)*len(self.axe_y))
            # computing spectrum variation on each mode
            var_spec = np.empty((len(self.modes)))
            for n in np.arange(len(self.modes)):
                if isinstance(self.modes[n], ScalarField):
                    data = self.modes[n].values
                    var_spec[n] = get_field_spec_std(data)
                elif isinstance(self.modes[n], VectorField):
                    datax = self.modes[n].comp_x
                    datay = self.modes[n].comp_y
                    stdx = get_field_spec_std(datax)
                    stdy = get_field_spec_std(datay)
                    var_spec[n] = (stdx + stdy)/2.
                else:
                    raise Exception()
            # normalize with min and max kappa
            var_spec = (var_spec - min_std)/(max_std - min_std)
            # storing in object
            self._ModalFields__spat_coh = var_spec
        if raw:
            return var_spec
        else:
            prof = Profile(np.arange(len(self.modes)), var_spec, unit_x='',
                           unit_y='', name="")
            return prof

    def display(self):
        # modes
        fig1 = plt.figure()
        modes = self.modes_as_tf
        if hasattr(modes, "magnitude"):
            plot1 = modes.display('magnitude')
        else:
            plot1 = modes.display()
        plt.title('POD modes')
        # temporal evolutions
        fig2 = plt.figure()
        tmp_x = [prof.x for prof in self.temp_evo]
        tmp_y = [prof.y for prof in self.temp_evo]
        plot2 = pplt.Displayer(tmp_x, tmp_y, data_type='profile', color='k')
        bm = pplt.ButtonManager(plot2)
        plot2.button_manager.link_to_other_graph(plot1)
        plt.title('Modes temporal evolution')
        # temporal evolution spectrum
        fig3 = plt.figure()
        specs = [prof.get_spectrum() for prof in self.temp_evo]
        tmp_x = [prof.x[1::] for prof in specs]
        tmp_y = [prof.y[1::] for prof in specs]
        plot3 = pplt.Displayer(tmp_x, tmp_y, data_type='profile',
                               kind='loglog', color='k')
        bm = pplt.ButtonManager(plot3)
        plot3.button_manager.link_to_other_graph(plot2)
        plt.title('Modes temporal evolution spectra')
        # cum nrj
        fig4 = plt.figure()
        cum_nrj = self.get_modes_energy(cum=True)
        tmp_x = [[x] for x in cum_nrj.x]
        tmp_y = [[y] for y in cum_nrj.y]
        plot4 = pplt.Displayer(tmp_x, tmp_y, data_type='profile', kind='plot',
                               marker='o', mfc='k', ls='none')
        bm = pplt.ButtonManager(plot4)
        cum_nrj.display(color='k', ls='-', marker=None)
        plt.xlim(xmin=0)
        plt.ylim(0, 1)
        plot4.button_manager.link_to_other_graph(plot2)
        plt.title('Modes cumulative energy')
        return plot1, plot2, plot3, plot4

    def display_recap(self, figsize=(15, 10)):
        """
        Display some important diagram for the decomposition.
        """
        print("Obsolete !")
        if self.decomp_type in ['pod', 'bpod']:
            plt.figure(figsize=figsize)
            plt.subplot(2, 3, 1)
            p_vals = self.get_spatial_coherence()
            p_vals.display(color='k')
            plt.axhline(0, color='k', linestyle=":")
            plt.title('Coherence indicator')
            plt.xlabel('Modes')
            plt.ylabel('Coherence')
            plt.subplot(2, 3, 2)
            self.modes[0].display()
            plt.title("Mode 1")
            plt.subplot(2, 3, 4)
            tmp_prof = self.get_modes_energy(cum=True)
            modes_nrj = self.get_modes_energy(raw=True)
            tmp_prof.y /= np.sum(modes_nrj)
            tmp_prof.display(color='k')
            plt.title('Cumulative modes energy')
            plt.ylim(ymin=0, ymax=1)
            plt.subplot(2, 3, 5)
            self.temp_evo[0].display(color='k')
            plt.title("Temporal evolution of mode 1")
            plt.subplot(2, 3, 3)
            self.modes[1].display()
            plt.title("Mode 2")
            plt.subplot(2, 3, 6)
            self.temp_evo[1].display(color='k')
            plt.title("Temporal evolution of mode 2")
        elif self.decomp_type == 'dmd':
            self.pulsation.change_unit('y', 'rad/s')
            self.growth_rate.change_unit('y', '1/s')
            plt.figure(figsize=figsize)
            plt.subplot(2, 3, 1)
            plt.plot(np.real(self.ritz_vals.y), np.imag(self.ritz_vals.y),
                     'ko')
            plt.title("Ritz eigenvalues in the complexe plane")
            plt.xlabel("Real part of Ritz eigenvalue")
            plt.ylabel("Imaginary part of Ritz eigenvalue")
            plt.xlim(-1, 1)
            plt.ylim(-1, 1)
            plt.subplot(2, 3, 2)
            plt.plot(self.pulsation.y, self.growth_rate.y, 'ko')
            plt.title("Growth rate spectrum")
            plt.xlabel("Pulsation [rad/s]")
            plt.ylabel("Growth rate [1/s]")
            x_max = np.max([np.abs(plt.xlim()[0]), np.abs(plt.xlim()[1])])
            plt.xlim(-x_max, x_max)
            plt.subplot(2, 3, 3)
            sorted_omega = np.sort(self.pulsation.y)
            delta_omega = np.mean(np.abs(sorted_omega[1::] -
                                  sorted_omega[0:-1:]))
            width = delta_omega/2.
            plt.bar(self.pulsation.y - width/2., self.mode_norms.y,
                    width=width)
            plt.title("Mode amplitude spectrum")
            plt.xlabel("Pulsation [rad/s]")
            plt.ylabel("Mode amplitude []")
            plt.subplot(2, 3, 5)
            stab_sort = np.argsort(np.abs(self.growth_rate.y))
            tmp_sf = self.modes[stab_sort[0]].copy()
            if isinstance(tmp_sf, ScalarField):
                tmp_sf.values = np.real(tmp_sf.values)
            elif isinstance(tmp_sf, VectorField):
                tmp_sf.comp_x = np.real(tmp_sf.comp_x)
                tmp_sf.comp_y = np.real(tmp_sf.comp_y)
            tmp_sf.display()
            plt.title("More stable mode (pulsation={:.2f})\n"
                      "(Real representation)"
                      .format(self.pulsation.y[stab_sort[-0]]))
            plt.subplot(2, 3, 4)
            tmp_prof = self.get_modes_energy()
            tmp_prof.y /= np.sum(tmp_prof.y)
            plt.plot(self.pulsation.y, tmp_prof.y, 'ko')
            plt.title('Modes energy')
            plt.ylabel("Normalized kinetic energy []")
            plt.xlabel("Pulsation [rad/s]")
            plt.ylim(ymin=0)
            plt.subplot(2, 3, 6)
            norm_sort = np.argsort(self.mode_norms.y)
            tmp_sf = self.modes[norm_sort[-1]].copy()
            if isinstance(tmp_sf, ScalarField):
                tmp_sf.values = np.real(tmp_sf.values)
            elif isinstance(tmp_sf, VectorField):
                tmp_sf.comp_x = np.real(tmp_sf.comp_x)
                tmp_sf.comp_y = np.real(tmp_sf.comp_y)
            tmp_sf.display()
            plt.title("Mode with the bigger norm (pulsation={:.2f})\n"
                      "(Real representation)"
                      .format(self.pulsation.y[norm_sort[-1]]))
        plt.tight_layout()


def _tsf_to_POD(tsf):
    # getting datas
    props = {}
    tsf = tsf.copy()
    tsf.crop_masked_border(inplace=True)
    tsf.fill(tof='spatial', kind='linear', inplace=True)
    ind_fields = np.arange(len(tsf.fields))
    props['ind_fields'] = ind_fields
    props['f_shape'] = tsf.fields[0].shape
    props['mean_field'] = tsf.get_mean_field()
    tsf = tsf.get_fluctuant_fields()
    props['super_mask'] = np.sum(tsf.mask, axis=0) == tsf.mask.shape[0]
    props['axe_x'], props['axe_y'] = tsf.axe_x, tsf.axe_y
    props['times'] = tsf.times
    props['unit_x'], props['unit_y'] = tsf.unit_x, tsf.unit_y
    props['unit_values'] = tsf.unit_values
    props['unit_times'] = tsf.unit_times
    # Link data
    values = [tsf.fields[t].values for t in ind_fields]
    del tsf
    snaps = [modred.VecHandleInMemory(values[t])
             for t in ind_fields]
    del values
    # return
    return snaps, props


def _tvf_to_POD(tvf):
    # getting datas
    props = {}
    tvf = tvf.copy()
    tvf.crop_masked_border(inplace=True)
    tvf.fill(tof='spatial', kind='linear', inplace=True)
    ind_fields = np.arange(len(tvf.fields))
    props['ind_fields'] = ind_fields
    props['f_shape'] = tvf.fields[0].shape
    props['mean_field'] = tvf.get_mean_field()
    tvf = tvf.get_fluctuant_fields()
    props['super_mask'] = np.sum(tvf.mask, axis=0) == tvf.mask.shape[0]
    props['axe_x'], props['axe_y'] = tvf.axe_x, tvf.axe_y
    props['times'] = tvf.times
    props['unit_x'], props['unit_y'] = tvf.unit_x, tvf.unit_y
    props['unit_values'] = tvf.unit_values
    props['unit_times'] = tvf.unit_times
    # Linking to snaps
    values = [[tvf.fields[t].comp_x, tvf.fields[t].comp_y]
              for t in ind_fields]
    del tvf
    snaps = [modred.VecHandleInMemory(np.transpose(values[i], (1, 2, 0)))
             for i in ind_fields]
    del values
    # return
    return snaps, props


def _POD_to_tsf(modes, props):
    modes_f = []
    for i in np.arange(len(modes)):
        tmp_field = ScalarField()
        tmp_field.import_from_arrays(props['axe_x'], props['axe_y'],
                                     modes[i].get(),
                                     mask=props['super_mask'],
                                     unit_x=props['unit_x'],
                                     unit_y=props['unit_y'],
                                     unit_values=props['unit_values'])
        modes[i] = 0
        modes_f.append(tmp_field)
    return modes_f


def _POD_to_tvf(modes, props):
    locals().update(props)
    modes_f = []
    for i in np.arange(len(modes)):
        tmp_field = VectorField()
        comp_x = modes[i].get()[:, :, 0]
        comp_y = modes[i].get()[:, :, 1]
        modes[i] = 0
        tmp_field.import_from_arrays(props['axe_x'], props['axe_y'],
                                     comp_x, comp_y,
                                     mask=props['super_mask'],
                                     unit_x=props['unit_x'],
                                     unit_y=props['unit_y'],
                                     unit_values=props['unit_values'])
        modes[i] = 0
        modes_f.append(tmp_field)
    return modes_f


def modal_decomposition(obj, kind='pod', obj2=None, wanted_modes='all',
                        max_vecs_per_node=1000, verbose=True):
    """
    Compute POD modes of the given fields using the snapshot method.

    Parameters
    ----------
    obj : TemporalFields, Profile or Points object
        Fields to extract modes from
    obj2 : same as obj
        Only used as second dataset for BPOD
    kind : string, optional
        Kind of decomposition, can be 'pod' (default), 'bpod' or 'dmd'.
    wanted_modes : string or number or array of numbers
        If 'all', extract all modes,
        If a number, extract first modes,
        If an array, extract the associated modes.
    max_vecs_per_node : integer, optional
        Number of fields that can be charged in memory.
        (More is faster but can lead to MemoryError)
    verbose : boolean, optional
        If 'True', display information.

    Returns
    -------
    modal_field : ModalField object
        .

    Notes
    -----
    You can use partially masked fields as input.
    If so, the asked values are lineary interpolated before doing the
    decomposition.
    """
    # check if modred is available
    if not MODRED:
        raise Exception("This feature need 'modred' to be installed")
    # Test parameters
    if not isinstance(obj, (TemporalFields)):
        raise TypeError()
    if kind == "bpod":
        if obj2 is None:
            raise ValueError()
        if not isinstance(obj2, obj.__class__):
            raise TypeError()
        if not obj2.shape == obj.shape:
            raise ValueError()
    if not isinstance(kind, STRINGTYPES):
        raise TypeError()
    if kind not in ['pod', 'bpod', 'dmd']:
        raise ValueError()
    if isinstance(wanted_modes, STRINGTYPES):
        if not wanted_modes == 'all':
            raise ValueError()
        wanted_modes = np.arange(len(obj.fields))
    elif isinstance(wanted_modes, NUMBERTYPES):
        wanted_modes = np.arange(wanted_modes)
    elif isinstance(wanted_modes, ARRAYTYPES):
        wanted_modes = np.array(wanted_modes)
        if wanted_modes.min() < 0 or wanted_modes.max() > len(obj.fields):
            raise ValueError()
    else:
        raise TypeError()
    try:
        max_vecs_per_node = int(max_vecs_per_node)
    except:
        raise TypeError()
    # getting datas
    if isinstance(obj, TemporalScalarFields):
        obj_type = 'TSF'
        snaps, props = _tsf_to_POD(obj)
        if kind == "bpod":
            snaps2, _ = _tsf_to_POD(obj2)
    elif isinstance(obj, TemporalVectorFields):
        obj_type = 'TVF'
        snaps, props = _tvf_to_POD(obj)
        if kind == "bpod":
            snaps2, _ = _tvf_to_POD(obj2)
    else:
        raise TypeError()
    globals().update(props)
    # Setting the decomposition mode
    eigvals = None
    eigvect = None
    ritz_vals = None
    mode_norms = None
    growth_rate = None
    pulsation = None
    if kind == 'pod':
        my_decomp = modred.PODHandles(np.vdot,
                                      max_vecs_per_node=max_vecs_per_node,
                                      verbosity=verbose)
        eigvals, eigvect = my_decomp.compute_decomp(snaps)
        del snaps
        wanted_modes = wanted_modes[wanted_modes < len(eigvals)]
        eigvect = np.array(eigvect)
        eigvect = eigvect[:, wanted_modes]
        eigvals = Profile(wanted_modes, eigvals[wanted_modes], mask=False,
                          unit_x=props['unit_times'], unit_y='')
    elif kind == 'bpod':
        my_decomp = modred.BPODHandles(np.vdot,
                                       max_vecs_per_node=max_vecs_per_node,
                                       verbosity=verbose)
        eigvect, eigvals, eigvect_l = my_decomp.compute_decomp(snaps, snaps2)
        del snaps
        del snaps2
        wanted_modes = wanted_modes[wanted_modes < len(eigvals)]
        eigvect = np.array(eigvect)
        eigvect = eigvect[:, wanted_modes]
        eigvals = Profile(wanted_modes, eigvals[wanted_modes], mask=False,
                          unit_x=props['unit_times'], unit_y='')
    elif kind == 'dmd':
        my_decomp = modred.DMDHandles(np.vdot,
                                      max_vecs_per_node=max_vecs_per_node,
                                      verbosity=verbose)
        ritz_vals, mode_norms, build_coeffs = my_decomp.compute_decomp(snaps)
        del snaps
        wanted_modes = wanted_modes[wanted_modes < len(ritz_vals)]
        # supplementary charac
        delta_t = props['times'][1] - props['times'][0]
        lambd_i = np.imag(ritz_vals)
        lambd_r = np.real(ritz_vals)
        lambd_mod = np.abs(ritz_vals)
        lambd_arg = np.zeros((len(ritz_vals)))
        mask = np.logical_and(lambd_i == 0, lambd_r <= 0)
        filt = np.logical_not(mask)
        lambd_arg[mask] = np.pi
        lambd_arg[filt] = 2*np.arctan(lambd_i[filt]/(lambd_r[filt] +
                                                     lambd_mod[filt]))
        sigma = np.log(lambd_mod)/delta_t
        omega = lambd_arg/delta_t
        # creating profiles
        unit_times = props['unit_times']
        ritz_vals = Profile(wanted_modes, ritz_vals[wanted_modes], mask=False,
                            unit_x=unit_times, unit_y='')
        mode_norms = Profile(wanted_modes, mode_norms[wanted_modes],
                             mask=False, unit_x=unit_times, unit_y='')
        growth_rate = Profile(wanted_modes, sigma[wanted_modes], mask=False,
                              unit_x=unit_times, unit_y=1./unit_times)
        pulsation = Profile(wanted_modes, omega[wanted_modes], mask=False,
                            unit_x=unit_times,
                            unit_y=make_unit('rad')/unit_times)
    else:
        raise ValueError("Unknown kind of decomposition : {}".format(kind))
    f_shape = props['f_shape']
    unit_values = props['unit_values']
    unit_times = props['unit_times']
    times = props['times']
    ind_fields = props['ind_fields']
    # Decomposing and getting modes
    if kind in ['pod', 'dmd']:
        modes = [modred.VecHandleInMemory(np.zeros(f_shape))
                 for i in np.arange(len(wanted_modes))]
        my_decomp.compute_modes(wanted_modes, modes)
    elif kind in ['bpod']:
        modes = [modred.VecHandleInMemory(np.zeros(f_shape))
                 for i in np.arange(len(wanted_modes))]
        adj_modes = [modred.VecHandleInMemory(np.zeros(f_shape))
                     for i in np.arange(len(wanted_modes))]
        my_decomp.compute_direct_modes(wanted_modes, modes)
        my_decomp.compute_adjoint_modes(wanted_modes, adj_modes)
    # Getting temporal evolution
    temporal_prof = []
    if kind in ['pod']:
        for n in np.arange(len(modes)):
            tmp_prof = eigvect[:, n]*eigvals.y[n]**.5
            tmp_prof = Profile(times, tmp_prof, mask=False,
                               unit_x=unit_times,
                               unit_y=unit_values)
            temporal_prof.append(tmp_prof)
    if kind in ['bpod']:
        for n in np.arange(len(modes)):
            tmp_prof = eigvect[:, n]*eigvals.y[n]**.5
            tmp_times = times[0:len(tmp_prof)]
            tmp_prof = Profile(tmp_times, tmp_prof, mask=False,
                               unit_x=unit_times,
                               unit_y=unit_values)
            temporal_prof.append(tmp_prof)
    elif kind == 'dmd':
        for n in np.arange(len(modes)):
            tmp_prof = np.real([ritz_vals.y[n]**(k) for k in ind_fields])
            tmp_prof = Profile(times, tmp_prof, mask=False, unit_x=unit_times,
                               unit_y=obj.unit_values)
            temporal_prof.append(tmp_prof)
    # Returning
    if obj_type == "TSF":
        modes_f = _POD_to_tsf(modes, props)
    elif obj_type == "TVF":
        modes_f = _POD_to_tvf(modes, props)
    del modes
    modal_field = ModalFields(kind, props['mean_field'], modes_f, wanted_modes,
                              temporal_prof,
                              eigvals=eigvals, eigvects=eigvect,
                              ritz_vals=ritz_vals, mode_norms=mode_norms,
                              growth_rate=growth_rate, pulsation=pulsation)
    return modal_field



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import matplotlib.pyplot as plt


def FDG_optimization(nodes, edges, edge_weights, node_values):
    try:
        import networkx as nx
    except ImportError:
        raise Exception("You need to install `networkx` to use this"
                        " functionality.")
    graph = nx.Graph()
    graph.add_nodes_from(nodes)
    for edge, weight in zip(edges, edge_weights):
        graph.add_edge(*edge, weight=weight)
    # set initial position
    init_pos = {node: (node_values[node], np.random.rand(1)) for node in nodes}
    node_pos = nx.fruchterman_reingold_layout(graph, dim=2, weight="weight",
                                              iterations=500, pos=init_pos)
    plt.figure()
    nx.draw(graph, with_labels=True, node_color=node_values, pos=init_pos)
    plt.figure()
    nx.draw(graph, with_labels=True, node_color=node_values, pos=node_pos)
    plt.figure()
    nx.draw_spring(graph)
    return node_pos



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.interpolate as spinterp
from ..utils.types import ARRAYTYPES, INTEGERTYPES, NUMBERTYPES, STRINGTYPES
from ..utils import make_unit
from . import scalarfield as sf
from . import profile as prof
from . import temporalscalarfields as tsf
from . import temporalfields as tf


class TemporalVectorFields(tf.TemporalFields):
    """
    Class representing a set of time-evolving velocity fields.
    """

    @property
    def Vx_as_sf(self):
        values = tsf.TemporalScalarFields()
        for i, field in enumerate(self.fields):
            values.add_field(field.comp_x_as_sf, time=self.times[i],
                             unit_times=self.unit_times)
        return values

    @property
    def Vx(self):
        dim = (len(self), self.shape[0], self.shape[1])
        values = np.empty(dim, dtype=float)
        for i, field in enumerate(self.fields):
            values[i, :, :] = field.comp_x[:, :]
        return values

    @property
    def Vy_as_sf(self):
        values = tsf.TemporalScalarFields()
        for i, field in enumerate(self.fields):
            values.add_field(field.comp_y_as_sf, time=self.times[i],
                             unit_times=self.unit_times)
        return values

    @property
    def Vy(self):
        dim = (len(self), self.shape[0], self.shape[1])
        values = np.empty(dim, dtype=float)
        for i, field in enumerate(self.fields):
            values[i, :, :] = field.comp_y[:, :]
        return values

    @property
    def magnitude_as_sf(self):
        values = tsf.TemporalScalarFields()
        for i, field in enumerate(self.fields):
            values.add_field(field.magnitude_as_sf, time=self.times[i],
                             unit_times=self.unit_times)
        return values

    @property
    def magnitude(self):
        dim = (len(self), self.shape[0], self.shape[1])
        values = np.empty(dim, dtype=float)
        for i, field in enumerate(self.fields):
            values[i, :, :] = field.magnitude[:, :]
        return values

    @property
    def theta_as_sf(self):
        values = tsf.TemporalScalarFields()
        for i, field in enumerate(self.fields):
            values.add_field(field.theta_as_sf, time=self.times[i],
                             unit_times=self.unit_times)
        return values

    @property
    def theta(self):
        dim = (len(self), self.shape[0], self.shape[1])
        values = np.empty(dim, dtype=float)
        for i, field in enumerate(self.fields):
            values[i, :, :] = field.theta[:, :]
        return values

    def get_time_auto_correlation(self):
        """
        Return auto correlation based on Vx and Vy.
        """
        Vx0 = self.fields[0].comp_x
        Vy0 = self.fields[0].comp_y
        norm = np.mean(Vx0*Vx0 + Vy0*Vy0)
        corr = np.zeros((len(self.times),))
        for i, time in enumerate(self.times):
            Vxi = self.fields[i].comp_x
            Vyi = self.fields[i].comp_y
            corr[i] = np.mean(Vx0*Vxi + Vy0*Vyi)/norm
        return prof.Profile(self.times, corr, mask=False,
                            unit_x=self.unit_times,
                            unit_y=make_unit(''))

    def get_mean_kinetic_energy(self):
        """
        Calculate the mean kinetic energy.
        """
        final_sf = sf.ScalarField()
        mean_vf = self.get_mean_field()
        values_x = mean_vf.comp_x_as_sf
        values_y = mean_vf.comp_y_as_sf
        final_sf = 1./2*(values_x**2 + values_y**2)
        return final_sf

    def get_tke(self):
        """
        Calculate the turbulent kinetic energy.
        """
        mean_field = self.get_mean_field()
        mean_x = mean_field.comp_x_as_sf
        mean_y = mean_field.comp_y_as_sf
        del mean_field
        tke = tsf.TemporalScalarFields()
        for i in np.arange(len(self.fields)):
            comp_x = self.fields[i].comp_x_as_sf - mean_x
            comp_y = self.fields[i].comp_y_as_sf - mean_y
            tke.add_field(1./2*(comp_x**2 + comp_y**2),
                          time=self.times[i],
                          unit_times=self.unit_times)
        return tke

    def get_turbulent_intensity(self):
        """
        Calculate the turbulent intensity.

        TI = sqrt(2/3*k)/sqrt(Vx**2 + Vy**2)
        """
        turb_int = (2./3.*self.get_tke())**(.5)/self.magnitude_as_sf
        return turb_int

    def get_mean_tke(self):
        tke = self.get_tke()
        mean_tke = tke.get_mean_field()
        return mean_tke

    def get_reynolds_stress(self, nmb_val_min=1):
        """
        Calculate the reynolds stress.

        Returns
        -------
        Re_xx, Re_yy, Re_xy : ScalarField objects
            Reynolds shear stress
        """
        # getting fluctuating velocities
        turb_vf = self.get_fluctuant_fields()
        u_p = turb_vf.Vx
        v_p = turb_vf.Vy
        mask = turb_vf.mask
        # rs_xx
        rs_xx = np.zeros(self.shape, dtype=float)
        rs_yy = np.zeros(self.shape, dtype=float)
        rs_xy = np.zeros(self.shape, dtype=float)
        mask_rs = np.zeros(self.shape, dtype=bool)
        # boucle sur les points du champ
        for i in np.arange(self.shape[0]):
            for j in np.arange(self.shape[1]):
                # boucle sur le nombre de champs
                nmb_val = 0
                for n in np.arange(len(turb_vf.fields)):
                    # check if masked
                    if not mask[n][i, j]:
                        rs_yy[i, j] += v_p[n][i, j]**2
                        rs_xx[i, j] += u_p[n][i, j]**2
                        rs_xy[i, j] += u_p[n][i, j]*v_p[n][i, j]
                        nmb_val += 1
                if nmb_val > nmb_val_min:
                    rs_xx[i, j] /= nmb_val
                    rs_yy[i, j] /= nmb_val
                    rs_xy[i, j] /= nmb_val
                else:
                    rs_xx[i, j] = 0
                    rs_yy[i, j] = 0
                    rs_xy[i, j] = 0
                    mask_rs[i, j] = True
        # masking and storing
        axe_x, axe_y = self.axe_x, self.axe_y
        unit_x, unit_y = self.unit_x, self.unit_y
        unit_values = self.unit_values
        rs_xx_sf = sf.ScalarField()
        rs_xx_sf.import_from_arrays(axe_x, axe_y, rs_xx, mask_rs,
                                    unit_x, unit_y, unit_values**2)
        rs_yy_sf = sf.ScalarField()
        rs_yy_sf.import_from_arrays(axe_x, axe_y, rs_yy, mask_rs,
                                    unit_x, unit_y, unit_values**2)
        rs_xy_sf = sf.ScalarField()
        rs_xy_sf.import_from_arrays(axe_x, axe_y, rs_xy, mask_rs,
                                    unit_x, unit_y, unit_values**2)
        return (rs_xx_sf, rs_yy_sf, rs_xy_sf)

    def get_spectral_filtering(self, fmin, fmax, order=2, inplace=False):
        """
        Perform a temporal spectral filtering

        Parameters
        ----------
        fmin, fmax : numbers
            Minimal and maximal frequencies
        order : integer, optional
            Butterworth filter order

        Returns
        -------
        filt_tvf : TemporalVectorFields
            Filtered temporal field
        """
        # prepare
        if inplace:
            tvf = self
        else:
            tvf = self.copy()
        # make spectral filtering on Vx and Vy
        ftsfx = self._get_comp_spectral_filtering('Vx', fmin=fmin,
                                                  fmax=fmax, order=order)
        ftsfy = self._get_comp_spectral_filtering('Vy', fmin=fmin,
                                                  fmax=fmax, order=order)
        for i in range(len(self)):
            tvf.fields[i].comp_x = ftsfx.fields[i].values
            tvf.fields[i].comp_y = ftsfy.fields[i].values
        # return
        if not inplace:
            return tvf

    def fill(self, tof='spatial', kind='linear', value=[0., 0.],
             inplace=False, crop=False):
        """
        Fill the masked part of the array in place.

        Parameters
        ----------
        tof : string
            Can be 'temporal' for temporal interpolation, or 'spatial' for
            spatial interpolation.
        kind : string, optional
            Type of algorithm used to fill.
            'value' : fill with a given value
            'nearest' : fill with nearest available data
            'linear' : fill using linear interpolation
            'cubic' : fill using cubic interpolation
        value : 2x1 array
            Value for filling, '[Vx, Vy]' (only usefull with tof='value')
        inplace : boolean, optional
            .
        crop : boolean, optional
            If 'True', TVF borders are croped before filling.
        """
        # TODO : utiliser prof.Profile.fill au lieu d'une nouvelle méthode de
        #        filling
        # checking parameters coherence
        if len(self.fields) < 3 and tof == 'temporal':
            raise ValueError("Not enough fields to fill with temporal"
                             " interpolation")
        if not isinstance(tof, STRINGTYPES):
            raise TypeError()
        if tof not in ['temporal', 'spatial']:
            raise ValueError()
        if not isinstance(kind, STRINGTYPES):
            raise TypeError()
        if kind not in ['value', 'nearest', 'linear', 'cubic']:
            raise ValueError()
        if isinstance(value, NUMBERTYPES):
            value = [value, value]
        elif not isinstance(value, ARRAYTYPES):
            raise TypeError()
        value = np.array(value)
        if crop:
            self.crop_masked_border(hard=False, inplace=True)
        # temporal interpolation
        if tof == 'temporal':
            # getting super mask (0 where no value are masked and where all
            # values are masked)
            masks = self.mask
            sum_masks = np.sum(masks, axis=0)
            super_mask = np.logical_and(0 < sum_masks,
                                        sum_masks < len(self.fields) - 2)
            # loop on each field position
            for i, j in np.argwhere(super_mask):
                # get time profiles
                prof_x = self.get_time_profile(component='comp_x', pt=[i, j],
                                               ind=True)
                prof_y = self.get_time_profile(component='comp_y', pt=[i, j],
                                               ind=True)
                # getting masked position on profile
                inds_masked_x = np.where(prof_x.mask)[0]
                inds_masked_y = np.where(prof_y.mask)[0]
                # creating interpolation function
                if kind == 'value':
                    def interp_x(x):
                        return value[0]
                    def interp_y(x):
                        return value[1]
                elif kind == 'nearest':
                    raise Exception("Not implemented yet")
                elif kind == 'linear':
                    prof_filt = np.logical_not(prof_x.mask)
                    interp_x = spinterp.interp1d(prof_x.x[prof_filt],
                                                 prof_x.y[prof_filt],
                                                 kind='linear')
                    prof_filt = np.logical_not(prof_y.mask)
                    interp_y = spinterp.interp1d(prof_y.x[prof_filt],
                                                 prof_y.y[prof_filt],
                                                 kind='linear')
                elif kind == 'cubic':
                    prof_filt = np.logical_not(prof_x.mask)
                    interp_x = spinterp.interp1d(prof_x.x[prof_filt],
                                                 prof_x.y[prof_filt],
                                                 kind='cubic')
                    prof_filt = np.logical_not(prof_y.mask)
                    interp_y = spinterp.interp1d(prof_y.x[prof_filt],
                                                 prof_y.y[prof_filt],
                                                 kind='cubic')
                else:
                    raise ValueError("Invalid value for 'kind'")
                # inplace or not
                fields = self.fields.copy()
                # loop on all x profile masked points
                for ind_masked in inds_masked_x:
                    try:
                        interp_val = interp_x(prof_x.x[ind_masked])
                    except ValueError:
                        continue
                    # putting interpolated value in the field
                    fields[ind_masked].comp_x[i, j] = interp_val
                    fields[ind_masked].mask[i, j] = False
                # loop on all y profile masked points
                for ind_masked in inds_masked_y:
                    try:
                        interp_val = interp_y(prof_y.x[ind_masked])
                    except ValueError:
                        continue
                    # putting interpolated value in the field
                    fields[ind_masked].comp_y[i, j] = interp_val
                    fields[ind_masked].mask[i, j] = False
        # spatial interpolation
        elif tof == 'spatial':
            if inplace:
                fields = self.fields
            else:
                tmp_tvf = self.copy()
                fields = tmp_tvf.fields
            for i, field in enumerate(fields):
                fields[i].fill(kind=kind, value=value, inplace=True)
        else:
            raise ValueError("Unknown parameter for 'tof' : {}".format(tof))
        # returning
        if inplace:
            self.fields = fields
        else:
            tmp_tvf = self.copy()
            tmp_tvf.fields = fields
            return tmp_tvf



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict
from matplotlib.collections import LineCollection
from os import path

import warnings
import matplotlib as mpl
import matplotlib.animation as mplani
import matplotlib.pyplot as plt
import numpy as np
import scipy.interpolate as spinterp
from matplotlib.widgets import Button, Slider


ARRAYTYPES = (np.ndarray, list, tuple)
NUMBERTYPES = (int, int, float, complex, np.float16, np.float32,
               np.float64, np.int16, np.int32, np.int64, np.int8)
STRINGTYPES = (str, str)


def is_sorted(data):
    return all(a <= b for a, b in zip(data[:-1], data[1:]))


def get_color_cycles():
    return mpl.rcParams['axes.prop_cycle'].by_key()['color']


def get_color_gradient(cmap='jet', number=10):
    """
    Return a gradient of color for plot uses.
    """
    cmap = plt.cm.get_cmap(name=cmap)
    colors = [cmap(i/(number - 1)) for i in range(number)]
    return colors


def make_discrete_cmap(interv_centers, cmap=None):
    """
    Create a discrete map, with intervals centered on the given values.
    """
    if cmap is None:
        cmap = plt.rcParams['image.cmap']
    cmap = plt.cm.get_cmap(cmap)
    # Get boundaries
    interv_bounds = [(interv_centers[i+1] + interv_centers[i])/2.
                     for i in range(len(interv_centers) - 1)]
    mini = interv_bounds[0] - abs(interv_bounds[1] - interv_bounds[0])
    maxi = interv_bounds[-1] + abs(interv_bounds[-1] - interv_bounds[-2])
    # Get color list
    color_list = []
    for i in range(len(interv_centers)):
        color_list += [cmap(i/float(len(interv_centers) - 1))]
    # Create new cmap
    new_cmap = mpl.colors.ListedColormap(color_list,
                                         name="Custom_discrete_map")
    norm = plt.Normalize(mini, maxi)
    # return
    return new_cmap, norm


def annotate_multiple(s, xy, xytext, *args, **kwargs):
    ans = []
    ax = plt.gca()
    an = ax.annotate(s, xy[0], xytext=xytext, *args, **kwargs)
    ans.append(an)
    d = {}
    try:
        d['xycoords'] = kwargs['xycoords']
    except KeyError:
        pass
    try:
        d['arrowprops'] = kwargs['arrowprops']
    except KeyError:
        pass
    for tmp_xy in xy[1:]:
        an = ax.annotate(s, tmp_xy, alpha=0.0, xytext=xytext, *args, **kwargs)
        ans.append(an)
    return ans


def mark_axe(txt, ax=None, loc=2, pad=0.3, borderpad=0., font_props=None,
             frameon=True):
    # get axe
    if ax is None:
        ax = plt.gca()
    # mark style
    if font_props is None:
        font_props = dict(fontweight='bold')
    # draw
    txt = mpl.offsetbox.AnchoredText(txt, loc=loc, prop=font_props, pad=pad,
                                     borderpad=borderpad,
                                     frameon=frameon)
    ax.add_artist(txt)
    return txt


def make_cmap(colors, position=None, name='my_cmap'):
    '''
    Return a color map cnstructed with the geiven colors and positions.

    Parameters
    ----------
    colors : Nx1 list of 3x1 tuple
        Each color wanted on the colormap. each value must be between 0 and 1.
    positions : Nx1 list of number, optional
        Relative position of each color on the colorbar. default is an
        uniform repartition of the given colors.
    name : string, optional
        Name for the color map
    '''
    # check
    if not isinstance(colors, ARRAYTYPES):
        raise TypeError()
    colors = np.array(colors, dtype=float)
    if colors.ndim != 2:
        raise ValueError()
    if colors.shape[1] != 3:
        raise ValueError()
    if position is None:
        position = np.linspace(0, 1, len(colors))
    else:
        position = np.array(position)
    if position.shape[0] != colors.shape[0]:
        raise ValueError()
    if not isinstance(name, STRINGTYPES):
        raise TypeError()
    # create colormap
    cdict = {'red': [], 'green': [], 'blue': []}
    for pos, color in zip(position, colors):
        cdict['red'].append((pos, color[0], color[0]))
        cdict['green'].append((pos, color[1], color[1]))
        cdict['blue'].append((pos, color[2], color[2]))
    cmap = mpl.colors.LinearSegmentedColormap(name, cdict, 256)
    # returning
    return cmap


class Formatter(mpl.ticker.ScalarFormatter):
    def __init__(self, order=0, fformat="%1.1f", offset=True, mathtext=True):
        self.oom = order
        self.fformat = fformat
        mpl.ticker.ScalarFormatter.__init__(self,
                                            useOffset=offset,
                                            useMathText=mathtext)

    def _set_orderOfMagnitude(self, nothing):
        self.orderOfMagnitude = self.oom

    def _set_format(self, vmin, vmax):
        self.format = self.fformat
        if self._useMathText:
            self.format = '${}$'.format(mpl.ticker._mathdefault(self.format))


def save_animation(animpath, fig=None, fields='all', writer='ffmpeg', fps=24,
                   title="", artist="IMTreatment", comment="",
                   bitrate=-1, codec='ffv1', dpi=150):
        """
        Save the current button manager displays as an animation.

        Parameters
        ----------
        animpath : string
            Path where to save animation
        fig : Figure instance
            Figure to save the animation from (if None, get the current one)
        fields : string or 2x1 list of numbers
            Fields interval to save. Default is 'all' for all the fields.
        writer : string
            Name of the writer to use
            (available writers are listed in
            'matplotlib.animation.writers.list()'
        codec : string
            One of the codec of the choosen writer (default to 'ffv1')
        fps : integer
            Number of frame per second (default to 24)
        bitrate : integer
            Video bitrate in kb/s (default to -1)
            Set this to -1 for letting the writter choose.
        dpi : integer
            dpi of the video images before compression (default to 150)
        title, artist, comment : strings
            Information added to the file metadata
        """
        fig = plt.gcf()
        try:
            bm = fig.button_manager
        except AttributeError:
            raise Exception("The current figure is not associated with a "
                            "button manager")
        bm.save_animation(animpath=animpath, fields=fields, writer=writer,
                          fps=fps, title=title, artist=artist, comment=comment,
                          bitrate=bitrate, codec=codec, dpi=dpi)


def use_perso_style():
    """
    Change matplotlib default style to something nicer
    """
    fp = path.dirname(__file__)
    plt.style.use(path.join(fp, r'perso.mplstyle'))
    # plt.rcParams["backend"] = "gtkcairo"


# Data manipulation:
def make_segments(x, y):
    '''
    Create list of line segments from x and y coordinates, in the correct
    format for LineCollection:
    an array of the form  numlines x (points per line) x 2 (x and y) array
    '''
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)
    return segments


# Interface to LineCollection:
def colored_plot(x, y, z=None, log='plot', min_colors=1000, colorbar=False,
                 color_label='', **kwargs):
    '''
    Plot a colored line with coordinates x and y

    Parameters
    ----------
    x, y : nx1 arrays of numbers
        coordinates of each points
    z : nx1 array of number, optional
        values for the color
    log : string, optional
        Type of axis, can be 'plot' (default), 'semilogx', 'semilogy',
        'loglog'
    min_colors : integer, optional
        Minimal number of different colors in the plot (default to 1000).
    colorbar : bool
        .
    color_label : string, optional
        Colorbar label if color is an array.
    kwargs : dict, optional
        list of arguments to pass to the common plot
        (see matplotlib documentation).
    '''
    # check parameters
    if not isinstance(x, ARRAYTYPES):
        raise TypeError()
    x = np.array(x)
    if not isinstance(y, ARRAYTYPES):
        raise TypeError()
    y = np.array(y)
    if len(x) != len(y):
        raise ValueError()
    if len(x) < 2:
        raise ValueError()
    length = len(x)
    if z is None:
        pass
    elif isinstance(z, ARRAYTYPES):
        if len(z) != length:
            raise ValueError()
        z = np.array(z)
    elif isinstance(z, NUMBERTYPES):
        z = np.array([z]*length)
    else:
        raise TypeError()
    if log not in ['plot', 'semilogx', 'semilogy', 'loglog']:
        raise ValueError()
    # classical plot if z is None
    if z is None:
        return plt.plot(x, y, **kwargs)
    # filtering nan values
    mask = np.logical_or(np.isnan(x), np.isnan(y))
    mask = np.logical_or(np.isnan(z), mask)
    filt = np.logical_not(mask)
    x = x[filt]
    y = y[filt]
    z = z[filt]
    length = len(x)
    # if length is too small, create artificial additional lines
    if length < min_colors:
        interp_x = spinterp.interp1d(np.linspace(0, 1, length), x)
        interp_y = spinterp.interp1d(np.linspace(0, 1, length), y)
        interp_z = spinterp.interp1d(np.linspace(0, 1, length), z)
        fact = np.ceil(min_colors/(length*1.))
        nmb_colors = length*fact
        x = interp_x(np.linspace(0., 1., nmb_colors))
        y = interp_y(np.linspace(0., 1., nmb_colors))
        z = interp_z(np.linspace(0., 1., nmb_colors))
    # make segments
    segments = make_segments(x, y)
    # make norm
    if 'norm' in list(kwargs.keys()):
        norm = kwargs.pop('norm')
    else:
        norm = plt.Normalize(np.min(z), np.max(z))
    # make cmap
    if 'cmap' in list(kwargs.keys()):
        cmap = kwargs.pop('cmap')
    else:
        cmap = plt.cm.__dict__[mpl.rc_params()['image.cmap']]
    # create line collection
    lc = LineCollection(segments, array=z, norm=norm, cmap=cmap, **kwargs)
    ax = plt.gca()
    ax.add_collection(lc)
    # adjuste og axis idf necessary
    if log in ['semilogx', 'loglog']:
        ax.set_xscale('log')
    if log in ['semilogy', 'loglog']:
        ax.set_yscale('log')
    plt.axis('auto')

    # colorbar
    if colorbar:
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        # fake up the array of the scalar mappable. Urgh...
        sm._A = []
        cb = plt.colorbar(sm)
        cb.set_label(color_label)
    return lc


# button gestion class
class ButtonManager(object):
    # TODO :
    #    -Axis limits have a werid behavior when using zoom

    def __init__(self, displayers, xlabel="", ylabel="", sharecb=True,
                 normcb=None, play_interval=2):
        # check if multiple displayers
        try:
            len(displayers)
        except TypeError:
            displayers = [displayers]
        # store Some more informations from displayers
        self.incr = 1
        self.ind = 0
        if len(displayers) == 0:
            raise ValueError()
        self.displayers = displayers
        self.ind_max = self._get_indmax_form_displs()
        for displ in self.displayers:
            displ.button_manager = self
        self.tmp_key_number = None
        self.sharecb = sharecb
        self.linked_graphs = []
        # check if a colorbar should be present
        is_cbs = [displ.mapped_colors for displ in displayers]
        self.is_cb = np.any(is_cbs)
        if self.is_cb:
            self.displ_cb = displayers[np.where(is_cbs)[0][0]]
        else:
            self.displ_cb = None
        # create figure or get the current one if compatible
        tmp_fig = plt.gcf()
        if len(tmp_fig.axes) == 0:
            self.fig = tmp_fig
        elif hasattr(tmp_fig, 'button_manager'):
            tmp_fig.button_manager.add_displayers(displayers)
            return None
        else:
            self.fig = plt.figure()
        # remove default keybinding
        self.fig.canvas.mpl_disconnect(
            self.fig.canvas.manager.key_press_handler_id)
        # associate the figure with the button manager
        self.fig.button_manager = self
        # Create the buttons and slider axes
        if self.is_cb:
            self.ax = mpl.axes.Axes(self.fig, [0.1, 0.2, .725, 0.75])
            self.fig.add_axes(self.ax)
            self.cbax = mpl.axes.Axes(self.fig, [0.85, 0.2, .025, 0.7])
            self.fig.add_axes(self.cbax)
        else:
            self.ax = mpl.axes.Axes(self.fig, [0.1, 0.2, .875, 0.75])
            self.fig.add_axes(self.ax)
            self.cbax = None
        self.axprev = mpl.axes.Axes(self.fig, [0.02, 0.02, 0.1, 0.05])
        self.fig.add_axes(self.axprev)
        self.axnext = mpl.axes.Axes(self.fig, [0.88, 0.02, 0.1, 0.05])
        self.fig.add_axes(self.axnext)
        self.axplay = mpl.axes.Axes(self.fig, [0.73, 0.02, 0.1, 0.05])
        self.fig.add_axes(self.axplay)
        self.axslid = mpl.axes.Axes(self.fig, [0.15, 0.02, 0.5, 0.05])
        self.fig.add_axes(self.axslid)
        # Add the button and slider and connect them
        self.button_kwargs = {"color": "w", "hovercolor": [.5]*3}
        self.slider_kwargs = {"facecolor": [.75]*3, "edgecolor": 'k', "lw": 1}
        self.slider_buff_kwargs = {'color': 'g', 'alpha': 0.5, 'lw': 0}
        self.slider_lims_kwargs = {'color': 'k', 'alpha': .5, 'lw': 0}
        self.bnext = Button(self.axnext, 'Next', **self.button_kwargs)
        self.bprev = Button(self.axprev, 'Previous', **self.button_kwargs)
        self.bplay = Button(self.axplay, 'Play', **self.button_kwargs)
        self.bnext.on_clicked(self.nextf)        # set up the apsect
        self.bprev.on_clicked(self.prevf)
        self.bplay.on_clicked(self.playf)
        self.bslid = Slider(self.axslid, "", valmin=1, valfmt='%d',
                            valmax=self.ind_max+1, valinit=1,
                            **self.slider_kwargs)
        self.bslid.on_changed(self.slid)
        self.slider_faces = {}
        # Add a timer for play
        self.on_play = False
        self.play_timer = self.fig.canvas.new_timer(interval=play_interval)
        self.play_timer.add_callback(self.timer_play)
        self.play_interval = play_interval
        self.could_interact = True
        self.ind_lims = [None, None]
        self.ind_lims_faces = [None, None]
        self.ind_lims_texts = [None, None]
        # add some keyboard shortcut
        self.fig.canvas.mpl_connect('key_press_event', self.keyf)
        # get aspect
        aspect, adjustable = self._get_aspect_from_displs()
        # Compute norm for colorbar, according to 'sharecb' and 'normcb'
        self.cb_norm = None
        self.cb = None
        if self.is_cb:
            cb_norm, cmap = self._get_cb_norm_from_displs(normcb=normcb)
            self.cb_norm = cb_norm
            self.cb = mpl.colorbar.ColorbarBase(self.cbax, norm=self.cb_norm,
                                                cmap=cmap)
            # self.cb.set_norm(self.cb_norm)
            self.cb.draw_all()
        # update with initial data
        self.update()
        # set texts and rescale
        self.xlim = [np.min([displ.xlim[0] for displ in self.displayers]),
                     np.max([displ.xlim[1] for displ in self.displayers])]
        self.ylim = [np.min([displ.ylim[0] for displ in self.displayers]),
                     np.max([displ.ylim[1] for displ in self.displayers])]
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.ax.set_xlabel(self.xlabel)
        self.ax.set_ylabel(self.ylabel)
        if aspect is not None:
            self.ax.set_aspect(aspect, adjustable=adjustable)
        self.ax.autoscale(False)
        if np.any([displ.data_type in ['field_1D', 'field_2D']
                   for displ in self.displayers]):
            self.ax.relim(visible_only=True)
            self.ax.autoscale_view(tight=True)        # set up the apsect
        else:
            self.ax.set_xlim(*self.xlim)
            self.ax.set_ylim(*self.ylim)
        # ensure that current ax is the good one
        plt.sca(self.ax)

    def _get_cb_norm_from_displs(self, normcb):
        # get cmap for colorbar
        try:
            cmap = self.displ_cb.dargs['cmap']
        except KeyError:
            cmap = mpl.rcParams['image.cmap']
        # secified norm
        if normcb is not None:
            cb_norm = normcb
        # no specified norm, but shared one
        elif self.sharecb:
            for displ in self.displayers:
                norm = displ.get_global_norm()
                displ.dargs['norm'] = norm
            cb_norm = self.displ_cb.dargs['norm']
        # different norm for each plot
        else:
            cb_norm = None
        # return
        cmap = plt.cm.get_cmap(cmap)
        return cb_norm, cmap

    def _get_aspect_from_displs(self):
        # get aspect
        aspects = []
        for displ in self.displayers:
            try:
                aspects.append(displ.dargs.pop('aspect'))
            except KeyError:
                pass
        if len(aspects) == 0:
            aspect = 'auto'
        elif np.any([asp == 'equal' for asp in aspects]):
            aspect = 'equal'
        elif np.any([asp == 'auto' for asp in aspects]):
            aspect = 'auto'
        else:
            raise ValueError()
        # get adjustable
        adjustables = []
        for displ in self.displayers:
            try:
                adjustables.append(displ.dargs.pop('adjustable'))
            except KeyError:
                pass
        if len(adjustables) == 0:
            adjustable = 'box'
        else:
            adjustable = adjustables[0]
        return aspect, adjustable

    def _get_indmax_form_displs(self):
        length = np.array([displ.length for displ in self.displayers])
        if np.any(length != self.displayers[0].length):
            raise ValueError()
        ind_max = self.displayers[0].length - 1
        return ind_max

    def prevf(self, event):
        # get increment
        key_number = self._get_key_number()
        if key_number is None:
            incr = self.incr
        else:
            incr = key_number
        #
        if self.ind == 0:
            return None
        # set new ind
        new_ind = self.ind - incr
        if new_ind > 0:
            self.ind = new_ind
        else:
            self.ind = 0
        # update
        self.update()

    def nextf(self, event):
        # get increment
        key_number = self._get_key_number()
        if key_number is None:
            incr = self.incr
        else:
            incr = key_number
        #
        if self.ind == self.ind_max:
            return None
        # set new ind
        new_ind = self.ind + incr
        if new_ind <= self.ind_max:
            self.ind = new_ind
        else:
            self.ind = self.ind_max
        # update
        self.update()

    def playf(self, event):
        # get and update time interval
        self._update_interv()
        self.play_timer.interval = self.play_interval
        #
        if self.on_play:
            self.play_timer.stop()
            self.on_play = False
            self.bplay.label.set_text('Play')
        else:
            self.play_timer.start()
            self.on_play = True
            self.bplay.label.set_text('Stop')

    def timer_play(self):
        # check if lims are set
        if self.ind_lims[1] is not None:
            ind1, ind2 = self.ind_lims
        else:
            ind1 = 0
            ind2 = self.ind_max + 1
        # update the display
        self.ind += self.incr
        if self.ind >= ind2:
            self.ind = ind1
        elif self.ind >= self.ind_max:
            self.ind = 0
        self.update()

    def goto(self):
        key_number = self._get_key_number()
        if key_number is None:
            return None
        if key_number > self.ind_max or key_number < 1:
            return None
        self.ind = key_number - 1
        self.update()

    def goto_end(self, event):
        key_number = self._get_key_number()
        if key_number is None:
            key_number = self.ind_max + 1
        if key_number > self.ind_max + 1:
            return None
        self.ind = key_number - 1
        self.update()

    def goto_beg(self, event):
        key_number = self._get_key_number()
        if key_number is None:
            key_number = 1
        if key_number > self.ind_max + 1:
            return None
        self.ind = key_number - 1
        self.update()

    def set_lims(self, lim1=None, lim2=None):
        # if lims are specified, change it
        if lim1 is not None or lim2 is not None:
            if lim1 is not None:
                if lim1 == 'del':
                    self.ind_lims[0] = None
                    self.ind_lims_faces[0].remove()
                    self.ind_lims_texts[0].remove()
                    self.ind_lims_faces[0] = None
                else:
                    self.ind_lims[0] = lim1
                    fac = self.axslid.axvspan(lim1, lim1+1,
                                              **self.slider_lims_kwargs)
                    self.ind_lims_faces[0] = fac
                    text = self.axslid.text(x=lim1+0.5, y=1.1, ha='center',
                                            s="{}".format(lim1 + 1))
                    self.ind_lims_texts[0] = text
            if lim2 is not None:
                if lim2 == 'del':
                    self.ind_lims[1] = None
                    self.ind_lims_faces[1].remove()
                    self.ind_lims_texts[1].remove()
                    self.ind_lims_faces[1] = None
                else:
                    self.ind_lims[1] = lim2
                    fac = self.axslid.axvspan(lim2, lim2+1,
                                              **self.slider_lims_kwargs)
                    self.ind_lims_faces[1] = fac
                    text = self.axslid.text(x=lim2+0.5, y=1.1, ha='center',
                                            s="{}".format(lim2 + 1))
                    self.ind_lims_texts[1] = text
                    # sort if necessary
                    if self.ind_lims[1] < self.ind_lims[0]:
                        self.ind_lims = self.ind_lims[::-1]
                        self.ind_lims_faces = self.ind_lims_faces[::-1]
                        self.ind_lims_texts = self.ind_lims_texts[::-1]
            return None
        # get ind
        key_nmb = self._get_key_number()
        if key_nmb is None:
            key_nmb = self.ind
        else:
            key_nmb -= 1
        # If ind is already a limite, remove it
        if key_nmb in self.ind_lims:
            if key_nmb == self.ind_lims[0]:
                self.set_lims(lim1='del')
            else:
                self.set_lims(lim2='del')
            return None
        # If lims are already set, remove them
        if self.ind_lims[1] is not None and self.ind_lims[0] is not None:
            self.set_lims(lim1='del', lim2='del')
        # If lims are not set, set the first one
        elif self.ind_lims[0] is None:
            self.set_lims(lim1=key_nmb)
        # Set the second limit
        else:
            self.set_lims(lim2=key_nmb)

    def goto_lims(self):
        self._get_key_number()
        if self.ind_lims[0] is None and self.ind_lims[1] is None:
            return None
        if self.ind_lims[0] is not None:
            if self.ind != self.ind_lims[0]:
                self.ind = self.ind_lims[0]
                self.update()
                return None
        if self.ind_lims[1] is not None:
            if self.ind != self.ind_lims[1]:
                self.ind = self.ind_lims[1]
                self.update()
                return None

    def keyf(self, event):
        # get directions
        if event.key in [' ', 'right', '+', 'l']:
            self.nextf(None)
        elif event.key in ['left', 'backspace', '-', 'h']:
            self.prevf(None)
        elif event.key in ['up', 'k']:
            self.goto_end(None)
        elif event.key in ['down', 'j']:
            self.goto_beg(None)
        elif event.key in ['enter', 'g']:
            self.goto()
        elif event.key in ['p', '.']:
            self.playf(None)
        elif event.key in ['i', '*']:
            self._update_incr()
        elif event.key in ['t', '/']:
            self._update_interv()
        elif event.key in ['q']:
            self.close()
        elif event.key in ['s']:
            self.save()
        elif event.key in ['pagedown', 'a']:
            self.set_lims()
        elif event.key in ['pageup', 'b']:
            self.goto_lims()
        else:
            pass
        # get numbers
        if event.key in ['{}'.format(i) for i in range(10)]:
            if self.tmp_key_number is None:
                self.tmp_key_number = int(event.key)
            else:
                self.tmp_key_number *= 10
                self.tmp_key_number += int(event.key)
        else:
            self.tmp_key_number = None

    def slid(self, event):
        # avoid recursion on update()
        if event - 1 == self.ind:
            return None
        self.ind = int(event) - 1
        self.update()

    def update(self):
        # deactivate buttons
        self.deactivate_buttons()
        # actualize slider
        self.bslid.set_val(self.ind + 1)
        # actualize all displayers
        for displ in self.displayers:
            displ.draw(self.ind, ax=self.ax, cb=False, remove_current=True,
                       rescale=False)
        if self.is_cb and not self.sharecb:
            norm = self.displ_cb.get_norm(self.ind)
            self.cb.set_norm(norm)
            self.cb.draw_all()
        # update linked graphs
        self._update_linked_graphs()
        # reactivate buttons
        self.activate_buttons()
        # update slide apparence
        self._update_slider_faces()

    def _update_slider_faces(self):
        displ = self.displayers[0]
        if not displ.use_buffer:
            return None
        # get buffered indices
        buff_inds = np.where(displ.displ_saved_inds != 0)[0]
        old_span = []
        # remove old faces
        for ind in list(self.slider_faces.keys()):
            if ind not in buff_inds:
                face = self.slider_faces.pop(ind)
                old_span.append(face)
        # add new ones
        ind = self.ind
        if ind not in list(self.slider_faces.keys()):
            # reuse old spans if possible
            if len(old_span) != 0:
                fac = old_span.pop(0)
                xy = fac.xy
                xy[:, 0] += ind - xy[0, 0]
                fac.set_xy(xy)
            # create a new face
            else:
                fac = self.axslid.axvspan(ind, ind+1,
                                          **self.slider_buff_kwargs)
            self.slider_faces[ind] = fac

    def _update_linked_graphs(self):
        # update linked graphs
        if len(self.linked_graphs) != 0:
            for graph in self.linked_graphs:
                if graph.ind != self.ind:
                    graph.ind = self.ind
                    graph.incr = self.incr
                    graph.update()

    def _get_key_number(self):
        key_number = self.tmp_key_number
        self.tmp_key_number = None
        return key_number

    def _update_incr(self):
        key_number = self._get_key_number()
        if key_number is not None:
            self.incr = key_number

    def _update_interv(self):
        key_number = self._get_key_number()
        if key_number is not None:
            self.play_interval = key_number

    def deactivate_buttons(self):
        self.could_interact = False
        self.bnext.set_active(False)
        self.bprev.set_active(False)
        self.bslid.set_active(False)

    def activate_buttons(self):
        self.could_interact = True
        self.bnext.set_active(True)
        self.bprev.set_active(True)
        self.bslid.set_active(True)

    def add_displayers(self, displayers):
        for displ in displayers:
            self.displayers.append(displ)
            displ.button_manager = self
        self.update()

    def link_to_other_graph(self, graph):
        if isinstance(graph, ButtonManager):
            if graph not in self.linked_graphs:
                self.linked_graphs.append(graph)
                graph.link_to_other_graph(self)
        elif hasattr(graph, 'button_manager'):
            self.link_to_other_graph(graph.button_manager)
        else:
            raise TypeError()

    def save_animation(self, animpath, fields='all', writer='ffmpeg', fps=24,
                       title="", artist="IMTreatment", comment="",
                       bitrate=-1, codec='ffv1', dpi=150):
        """
        Save the button manager displays as an animation.

        Parameters
        ----------
        animpath : string
            Path where to save animation
        fields : string or 2x1 list of numbers
            Fields interval to save. Default is 'all' for all the fields.
        writer : string
            Name of the writer to use
            (available writers are listed in
            'matplotlib.animation.writers.list()'
        codec : string
            One of the codec of the choosen writer (default to 'ffv1')
        fps : integer
            Number of frame per second (default to 24)
        bitrate : integer
            Video bitrate in kb/s (default to -1)
            Set this to -1 for letting the writter choose.
        dpi : integer
            dpi of the video images before compression (default to 150)
        title, artist, comment : strings
            Information added to the file metadata
        """
        # Check
        if bitrate == "default":
            bitrate = -1
        # Get first and last field
        if fields == "all":
            fields = [0, self.ind_max]
        # Get writer
        try:
            Writer = mplani.writers[writer]
        except KeyError:
            raise ValueError("{} not available as writer, try one of these "
                             " {}".format(writer, mplani.writers.list()))

        metadata = dict(title=title, artist=artist,
                        comment=comment)
        writer = Writer(fps=fps, metadata=metadata, bitrate=bitrate,
                        codec=codec)
        # write
        backup_ind = self.ind
        self.ind = fields[0]
        self.update()
        with writer.saving(self.fig, animpath, dpi):
            for i in range(fields[1] - fields[0]):
                self.nextf(None)
                writer.grab_frame()
        # restore
        self.ind = backup_ind
        self.update()

    def close(self):
        if self.on_play:
            self.playf(None)
#        self.__del__()

    def __del__(self):
        pass
#        plt.close(self.fig)
#        for displ in self.displayers:
#            del displ
#        del self.displ_cb

    def save(self):
        pass
#        self.fig.canvas.toolbar.save_figure()


class Displayer(object):

    points_default_args = {"kind": "scatter"}
    profile_default_args = {"kind": "plot"}
    field_1D_default_args = {"kind": "matrix",
                             # "interpolation": "nearest",
                             "aspect": "equal"}
    field_2D_default_args = {"kind": "quiver",
                             "aspect": "equal"}

    def __init__(self, x, y, values=None, data_type=None, sharebds=True,
                 buffer_size=100, **kwargs):
        # get figure
        if "ax" not in list(kwargs.keys()):
            self.ax = None
            self.fig = None
        else:
            self.ax = kwargs.pop("ax")
            self.fig = self.ax.figure
        # get data
        self.x = np.asarray(x)
        self.y = np.asarray(y)
        if len(self.x) != 0:
            try:
                tmp_x = np.concatenate(self.x)
            except ValueError:
                tmp_x = self.x
            self.xlim = [np.min(tmp_x[~np.isnan(tmp_x)]),
                         np.max(tmp_x[~np.isnan(tmp_x)])]
            try:
                tmp_y = np.concatenate(self.y)
            except:
                tmp_y = self.y
            self.ylim = [np.min(tmp_y[~np.isnan(tmp_y)]),
                         np.max(tmp_y[~np.isnan(tmp_y)])]
        if values is None:
            self.values = None
        else:
            self.values = np.asarray(values)
        self.vmin = None
        self.vmax = None
        self.colors = None
        self.magnitude = None
        self.sharebds = sharebds
        # check if data is multidimensionnal
        self.multidim = False
        try:
            x[0][0]
            self.multidim = True
        except IndexError:
            pass
        if isinstance(x[0], list) and isinstance(y[0], list):
            self.multidim = True
        if self.multidim:
            self.length = len(self.x)
        else:
            self.length = 1
        self.curr_ind = 0
        # place to store the drawings
        self.draws = [None]*self.length
        self.curr_draw = None
        self.displ_saved_inds = np.zeros(self.length, dtype=int)
        self.displ_saved_curr_ind = 1
        if buffer_size is None:
            self.max_saved_displ = 0
            self.use_buffer = False
        else:
            self.max_saved_displ = buffer_size
            self.use_buffer = True
        # try to guess the data type
        tmp_x, tmp_y, tmp_values, tmp_colors, tmp_magn = self.get_data(i=0)
        if data_type is None:
            if values is None:
                if is_sorted(tmp_x) or is_sorted(tmp_y):
                    self.data_type = "profile"
                else:
                    self.data_type = "points"
            else:
                if tmp_values.ndim == 1:
                    self.data_type = "points"
                elif tmp_values.ndim == 2:
                    self.data_type = "field_1D"
                elif tmp_values.ndim == 3:
                    self.data_type = "field_2D"
                else:
                    raise ValueError("Unable to detect the data type")
        elif data_type in ["points", "profile", "field_1D", "field_2D"]:
            self.data_type = data_type
        else:
            raise ValueError("Unknown 'data_type' argument")
        # set default values according to data type
        if self.data_type == "points":
            self.dargs = self.points_default_args.copy()
        elif self.data_type == "profile":
            self.dargs = self.profile_default_args.copy()
        elif self.data_type == "field_1D":
            self.dargs = self.field_1D_default_args.copy()
            if "color" in list(kwargs.keys()):
                if self.multidim:
                    self.colors = [kwargs.pop('color')]*self.length
                else:
                    self.colors = kwargs.pop('color')
            else:
                self.colors = self.values
        elif self.data_type == "field_2D":
            self.dargs = self.field_2D_default_args.copy()
            if self.multidim:
                self.magnitude = [(self.values[i][0]**2 +
                                   self.values[i][1]**2)**.5
                                  for i in range(self.length)]
            else:
                self.magnitude = (self.values[0]**2 +
                                  self.values[1]**2)**.5
            if "color" in list(kwargs.keys()):
                if self.multidim:
                    self.colors = [kwargs.pop('color')]*self.length
                else:
                    self.colors = kwargs.pop('color')
            else:
                self.colors = np.asarray(self.magnitude)
        # check if colors are mapped into data or not
        self.mapped_colors = False
        try:
            ndim = self.colors[0].ndim
            if self.colors[0].shape == self.values.shape[-ndim::]:
                self.mapped_colors = True
        except:
            pass
        # set user defined display arguments
        if 'kind' in list(kwargs.keys()):
            if kwargs['kind'] is None:
                kwargs.pop('kind')
        self.dargs.update(kwargs)

    def get_data(self, i=None):
        if not self.multidim:
            return self.x, self.y, self.values, self.colors, self.magnitude
        elif self.multidim and i is not None:
            if self.values is None:
                tmp_values = None
            else:
                tmp_values = self.values[i]
            if self.colors is None:
                tmp_colors = None
            else:
                tmp_colors = self.colors[i]
            if self.magnitude is None:
                tmp_magn = None
            else:
                tmp_magn = self.magnitude[i]
            return self.x[i], self.y[i], tmp_values, tmp_colors, tmp_magn
        else:
            raise ValueError()

    def get_data_at_point(self, x, y, i=None):
        if i is None:
            i = self.curr_ind
        # Get data
        tmp_x, tmp_y, tmp_values, tmp_colors, tmp_magn = self.get_data(i=i)
        dic = OrderedDict()
        # check if mouse is too far from points
        if (x > self.xlim[1] or x < self.xlim[0] or
                y > self.ylim[1] or y < self.ylim[0]):
            return None
        #
        ind_x = np.argmin(np.abs(x - tmp_x))
        # if a profile
        if self.data_type in ['profile', 'points']:
            dic['x'] = tmp_x[ind_x]
            dic['y'] = tmp_y[ind_x]
        elif self.data_type in ['field_1D']:
            ind_y = np.argmin(np.abs(y - tmp_y))
            dic['x'] = tmp_x[ind_x]
            dic['y'] = tmp_y[ind_y]
            dic['value'] = tmp_values[ind_x, ind_y]
        elif self.data_type in ['field_2D']:
            ind_y = np.argmin(np.abs(y - tmp_y))
            dic['x'] = tmp_x[ind_x]
            dic['y'] = tmp_y[ind_y]
            dic['Vx'] = tmp_values[0, ind_x, ind_y]
            dic['Vy'] = tmp_values[1, ind_x, ind_y]
        else:
            raise Exception()
        return dic

    def get_norm(self, i):
        vmin = np.min(self.colors[i])
        vmax = np.min(self.colors[i])
        return mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    def get_global_norm(self):
        if self.vmin is None:
            tmp_mins = []
            for col in self.colors:
                filt = ~np.isnan(col)
                if np.any(filt):
                    tmp_mins.append(np.min(col[filt]))
            self.vmin = np.min(tmp_mins)
        if self.vmax is None:
            tmp_maxs = []
            for col in self.colors:
                filt = ~np.isnan(col)
                if np.any(filt):
                    tmp_maxs.append(np.max(col[filt]))
            self.vmax = np.max(tmp_maxs)
        return mpl.colors.Normalize(vmin=self.vmin, vmax=self.vmax)

    def _toggle_visibility(self, obj):
        # if obj is an array of objects
        try:
            obj[0]
            for single_obj in obj:
                self._toggle_visibility(single_obj)
            return None
        except TypeError:
            pass
        # Remove or add obj according to its type
        # Classic type
        if isinstance(obj, (mpl.lines.Line2D, mpl.image.AxesImage,
                            mpl.collections.Collection)):
            try:
                obj.remove()
            except ValueError:
                self.ax.add_artist(obj)
        # Contour type
        elif isinstance(obj, mpl.contour.QuadContourSet):
            try:
                for tmp_obj in obj.collections:
                    tmp_obj.remove()
            except ValueError:
                for tmp_obj in obj.collections:
                    self.ax.add_artist(tmp_obj)
        # Streamplot type
        elif isinstance(obj, mpl.streamplot.StreamplotSet):
            try:
                obj.lines.remove()
                if not hasattr(obj.arrows, 'patches'):
                    obj.arrows.patches = self.ax.patches
                # TODO : memory leak
                #       (because arrows removing is not implemented)
                self.ax.patches = []
            except ValueError:
                self.ax.add_artist(obj.lines)
                self.ax.patches = obj.arrows.patches
        else:
            raise ValueError("{}".format(obj))

    def draw(self, i=None, ax=None, cb=False, remove_current=False,
             rescale=True):
        # Do nothing if I is too big...
        if i is not None:
            if i >= len(self.draws):
                warnings.warn("Indice too big")
                return self.draws[self.curr_ind]
        self.curr_ind = i
        # check data
        if self.multidim and i is None:
            raise ValueError()
        if i is None:
            i = 0
        # remove current draw
        if (self.curr_draw is not None and
                self.draws[self.curr_draw] is not None and
                remove_current):
            self._toggle_visibility(self.draws[self.curr_draw])
        if self.use_buffer:
            # if draw already computed
            if self.draws[i] is not None:
                self._toggle_visibility(self.draws[i])
                self.curr_draw = i
            # else make a new one
            else:
                self.curr_draw = i
                self.draws[i] = self.draw_new(i=i, ax=ax, cb=cb,
                                              rescale=rescale)
            # keep trace of saved displ
            # (deleting first ones if too much of them)
            self.displ_saved_inds[i] = self.displ_saved_curr_ind
            self.displ_saved_curr_ind += 1
            if np.sum(self.displ_saved_inds != 0) > self.max_saved_displ:
                # delete first displ
                ind = np.min(self.displ_saved_inds[self.displ_saved_inds != 0])
                ind = np.where(ind == self.displ_saved_inds)[0]
                if len(ind) != 0:
                    ind = ind[0]
                    self.draws[ind] = None
                    self.displ_saved_inds[ind] = 0.
            # returning
            return self.draws[i]
        else:
            # returning
            return self.draw_new(i=i, ax=ax, cb=cb, rescale=rescale)

    def draw_new(self, i=None, ax=None, cb=False, rescale=True):
        # check data
        if self.multidim and i is None:
            raise ValueError()
        if i is None:
            i = 0
        # draw
        if ax is None and self.ax is None:
            if self.fig is None:
                self.fig = plt.gcf()
            self.ax = self.fig.gca()
            ax = self.ax
        elif ax is None:
            ax = self.ax
        else:
            self.ax = ax
        tmp_x, tmp_y, tmp_values, tmp_colors, tmp_magn = self.get_data(i=i)
        # continue if there is nothing to draw
        if len(tmp_x) == 0:
            return None
        dargs = self.dargs.copy()
        kind = dargs.pop('kind')
        try:
            aspect = dargs.pop('aspect')
        except KeyError:
            aspect = None
        try:
            adjustable = dargs.pop('adjustable')
        except KeyError:
            adjustable = 'box'
        #
        if kind == 'scatter':
            if tmp_values is None:
                plot = ax.scatter(tmp_x, tmp_y, **dargs)
            else:
                if 'c' not in dargs.keys() and 'color' not in dargs.keys():
                    dargs['c'] = tmp_values
                plot = ax.scatter(tmp_x, tmp_y, **dargs)
        elif kind == 'plot':
            plot = ax.plot(tmp_x, tmp_y, **dargs)
        elif kind == 'colored_plot':
            plot = ax.plot = colored_plot(tmp_x, tmp_y, z=tmp_values, **dargs)
        elif kind == 'semilogx':
            tmp_filt = ~np.isnan(tmp_x)
            tmp_filt = np.logical_or(tmp_filt, tmp_x < 0)
            tmp_x[~tmp_filt] = np.nan
            plot = ax.semilogx(tmp_x, tmp_y, **dargs)
        elif kind == 'semilogy':
            tmp_filt = ~np.isnan(tmp_y)
            tmp_filt = np.logical_or(tmp_filt, tmp_y < 0)
            tmp_y[~tmp_filt] = np.nan
            plot = ax.semilogy(tmp_x, tmp_y, **dargs)
        elif kind == 'loglog':
            tmp_filt = ~np.isnan(tmp_x)
            tmp_filt = np.logical_or(tmp_filt, tmp_x < 0)
            tmp_x[~tmp_filt] = np.nan
            tmp_filt = ~np.isnan(tmp_y)
            tmp_filt = np.logical_or(tmp_filt, tmp_y < 0)
            tmp_y[~tmp_filt] = np.nan
            plot = ax.loglog(tmp_x, tmp_y, **dargs)
        elif kind == 'matrix':
            delta_x = tmp_x[1] - tmp_x[0]
            delta_y = tmp_y[1] - tmp_y[0]
            plot = ax.imshow(tmp_values.transpose(),
                             extent=(tmp_x[0] - delta_x/2.,
                                     tmp_x[-1] + delta_x/2.,
                                     tmp_y[0] - delta_y/2.,
                                     tmp_y[-1] + delta_y/2.),
                             origin='lower', **dargs)
        elif kind == "contour":
            plot = ax.contour(tmp_x, tmp_y, tmp_values.transpose(), **dargs)
        elif kind == "contourf":
            plot = ax.contourf(tmp_x, tmp_y, tmp_values.transpose(), **dargs)
        elif kind == "quiver":
            if 'color' in list(dargs.keys()):
                C = dargs.pop('color')
                if 'c' in dargs.keys():
                    dargs.pop('c')
            else:
                C = tmp_magn
                if 'c' in dargs.keys():
                    dargs.pop('c')
            plot = ax.quiver(tmp_x, tmp_y, tmp_values[0].transpose(),
                             tmp_values[1].transpose(), C.transpose(),
                             **dargs)
        elif kind == "stream":
            # set adptative linewidth
            if 'lw' in list(dargs.keys()):
                tmp_lw = dargs.pop('lw')
            elif 'linewidth' in list(dargs.keys()):
                tmp_lw = dargs.pop('linewidth')
            else:
                tmp_lw = 1
            if np.array(tmp_lw).shape != ():
                pass
            else:
                tmp_magn[np.isnan(tmp_magn)] = 0
                tmp_lw *= 0.1 + 0.9*tmp_magn/np.max(tmp_magn)
                tmp_lw = tmp_lw.transpose()
            # set color
            if np.array(tmp_colors).shape != ():
                tmp_colors = tmp_colors.transpose()
            # plot
            Vx = tmp_values[0].transpose()
            Vy = tmp_values[1].transpose()
            plot = ax.streamplot(tmp_x, tmp_y, Vx, Vy, color=tmp_colors,
                                 linewidth=tmp_lw, **dargs)
        else:
            raise Exception("Unknown kind of plot : {}".format(kind))
        if aspect is not None:
            try:
                self.ax.set_aspect(aspect, adjustable=adjustable)
            except:
                pass
        if rescale:
            if self.data_type in ['field_1D', 'field_2D']:
                self.ax.relim(visible_only=True)
                self.ax.autoscale_view(tight=True)        # set up the apsect
                self.ax.set_xlim(*self.xlim)
                self.ax.set_ylim(*self.ylim)
            else:
                # self.ax.relim(visible_only=True)
                self.ax.autoscale_view(tight=True)        # set up the apsect
                # self.ax.set_xlim(*self.xlim)
                # self.ax.set_ylim(*self.ylim)
        return plot

    def draw_multiple(self, inds, sharecb=False, sharex=False, sharey=False,
                      ncol=None, nrow=None):
        nmb_fields = len(inds)
        # get figure
        fig = plt.gcf()
        if len(fig.axes) == 0 or len(fig.axes) != len(inds):
            # creating new figure with subplots
            if ncol is None:
                ncol = int(np.sqrt(nmb_fields))
            if nrow is None:
                nrow = int(np.ceil(float(nmb_fields)/ncol))
            if ncol*nrow < len(inds):
                raise ValueError()
            # creating axes
            fig, axs = plt.subplots(nrows=nrow, ncols=ncol, sharex=sharex,
                                    sharey=sharey)
        elif len(fig.axes) == len(inds):
            # reuse the current axes
            axs = np.array(fig.axes)
        # getting min and max
        if sharecb:
            if 'norm' not in list(self.dargs.keys()):
                if "vmin" not in list(self.dargs.keys()):
                    vmin = np.min([np.min(self.colors[ind]) for ind in inds])
                else:
                    vmin = self.dargs.pop('vmin')
                if "vmax" not in list(self.dargs.keys()):
                    vmax = np.max([np.max(self.colors[ind]) for ind in inds])
                else:
                    vmax = self.dargs.pop('vmax')
                norm = plt.Normalize(vmin=vmin, vmax=vmax)
                self.dargs['norm'] = norm
        # displaying the wanted fields
        for i, ind in enumerate(inds):
            ax = axs.flat[i]
            plt.sca(ax)
            self.draw(ind, ax=ax, rescale=True, remove_current=False)
            DataCursorTextDisplayer(self, i=ind)
        # deleting the non-wanted axes
        for ax in axs.flat[nmb_fields::]:
            fig.delaxes(ax)
        plt.tight_layout()
        # same colorbar
        if sharecb:
            vmin = self.dargs['norm'].vmin
            vmax = self.dargs['norm'].vmax
            norm = plt.Normalize(vmin=vmin, vmax=vmax)
            fig.subplots_adjust(right=0.8)
            cbar_ax = fig.add_axes([0.85, 0.05, .025, 0.925])
            mpl.colorbar.ColorbarBase(cbar_ax, norm=norm,
                                      orientation='vertical')
            plt.tight_layout(rect=[0., 0., 0.85, 1.])
        # ensure correct ax for pyplot
        plt.sca(axs.flatten()[nmb_fields - 1])
        # returning
        return fig, axs


class DataCursorTextDisplayer(object):

    def __init__(self, displayer, i=None, precision=3):
        if i is None:
            self.displ_ind = None
        else:
            self.displ_ind = i
        self.displayer = displayer
        self.displayer.ax.format_coord = self.__default_formatter
        self.precision = precision

    def __default_formatter(self, x, y):
        data = self.displayer.get_data_at_point(x, y, i=self.displ_ind)
        if data is None:
            return ''
        text = ''
        for key in list(data.keys()):
            text += "{}=".format(key)
            text += self.__number_formatter(data[key])
            text += "     "
        return text

    def __number_formatter(self, number):
        # if string
        try:
            number.format()
            return number
        except AttributeError:
            pass
        if int(number) == number:
            return "{}".format(int(number))
        else:
            return "{:.4g}".format(number)

    def __get_data_from_ax(self, x, y):
        return self.displayer.get_data_at_point(x, y)


class DataCursorPoints(object):

    def __init__(self, ax, tolerance=5, offsets=(-20, 20),
                 formatter=None, display_all=False,
                 color=(0.76, 0.86, 0.92)):
        """
        A simple data cursor widget that displays the x,y location of a
        matplotlib artist when it is selected.

        Parameters
        ----------
        artists : sequence of matplotlib Artists
            is the matplotlib artist or sequence of artists that will be
            selected.
        tolerance : integer
            is the radius (in points) that the mouse click must be
            within to select the artist.
        offsets : 2x1 tuple of integer
            is a tuple  of (x,y) offsets in points from the selected
            point to the displayed annotation box
        formatter : function
            is a callback function which takes 2 numeric arguments and
            returns a string
        display_all : boolean
            controls whether more than one annotation box will
            be shown if there are multiple axes.  Only one will be shown
            per-axis, regardless.
        color : matplotlib color
            color of the information box

        Notes:
        ------
        Credit to
        http://stackoverflow.com/a/4674445/190597 (Joe Kington)
        http://stackoverflow.com/a/20637433/190597 (unutbu)
        """
        raise Exception("Not (properly) implemented yet")
        # self._points = np.column_stack((x, y))
        if formatter is None:
            self.formatter = self._default_fmt
        else:
            self.formatter = formatter
        self.offsets = offsets
        self.display_all = display_all
        # try:
        #     artists[0]
        # except TypeError:
        #     artists = [artists]
        # self.artists = artists
        self.axes = tuple(set(art.axes for art in self.artists))
        self.figures = tuple(set(ax.figure for ax in self.axes))
        self.current_displayed_xy = []
        self.annotations = {}
        self.color = color
        for ax in self.axes:
            self.annotations[ax] = self.annotate(ax)
        for artist in self.artists:
            artist.set_picker(tolerance)
        for fig in self.figures:
            fig.canvas.mpl_connect('pick_event', self)

    def annotate(self, ax):
        """
        Draws and hides the annotation box for the given axis "ax".
        """
        annotation = ax.annotate(self.formatter, xy=(0, 0), ha='right',
                                 xytext=self.offsets,
                                 textcoords='offset points', va='bottom',
                                 bbox=dict(boxstyle='round,pad=0.5',
                                           fc=self.color, alpha=1.),
                                 arrowprops=dict(arrowstyle='->',
                                                 connectionstyle='arc3,rad=0'))
        annotation.set_visible(False)
        return annotation

    def snap(self, x, y):
        """
        Return the value in self._points closest to (x, y).
        """
        idx = np.nanargmin(((self._points - (x, y))**2).sum(axis=-1))
        return self._points[idx]

    def _default_fmt(self, x, y):
        return 'x: {x:0.2f}\ny: {y:0.2f}'.format(x=x, y=y)

    def get_color_from_event(self, event):
        # get color from scatterplots
        color = None
        try:
            x, y = event.mouseevent.xdata, event.mouseevent.ydata
            x, y = self.snap(x, y)
            colors = event.artist.get_facecolor()
            color = colors[np.logical_and(self._points[:, 0] == x,
                                          self._points[:, 1] == y)][0]
        except:
            pass
        # get color from classic plots
        try:
            color = event.artist.get_color()
            color = mpl.colors.colorConverter.to_rgba_array(color)[0]
        except:
            pass
        # return
        return color

    def __call__(self, event):
        """
        Intended to be called through "mpl_connect".
        """
        x, y = event.mouseevent.xdata, event.mouseevent.ydata
        x, y = self.snap(x, y)
        # get point color to color annotation
        annotation = self.annotations[event.artist.axes]
        color = self.get_color_from_event(event)
        color = np.array(color)
        color = color + 0.75*(1 - color)
        annotation.set_backgroundcolor(color)
        if x is not None:
            if not self.display_all:
                # Hide any other annotation boxes...
                for ann in list(self.annotations.values()):
                    ann.set_visible(False)
            # Update the annotation in the current axis..
            # if already annotated, remove the annotation
            if [x, y] in self.current_displayed_xy:
                annotation.set_visible(False)
                self.current_displayed_xy.remove([x, y])
            else:
                annotation.xy = x, y
                annotation.set_text(self.formatter(x, y))
                annotation.set_visible(True)
                self.current_displayed_xy.append([x, y])
            event.canvas.draw()



# -*- coding: utf-8 -*-
#!/bin/env python3

# Copyright (C) 2003-2007 Gaby Launay

# Author: Gaby Launay  <gaby.launay@tutanota.com>
# URL: https://framagit.org/gabylaunay/IMTreatment

# This file is part of IMTreatment.

# IMTreatment is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.

# IMTreatment is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
try:
    dirname = os.path.dirname(os.path.realpath(__file__))
    sys.path.append(dirname)
    os.chdir(dirname)
except:
    pass
import matplotlib as mpl
mpl.use('Agg')

import numpy as np

import pytest
from helper import sane_parameters, parametric_test

import matplotlib.pyplot as plt
import unum

from IMTreatment import (field_treatment as imtft, file_operation as imtio,
                         make_unit, TemporalVectorFields, TemporalScalarFields)
from IMTreatment.core import TemporalVectorFields


class TestFileOperation(object):
    """ Done """

    def setup(self):
        sane_parameters()
        self.SF1_nomask = imtio.import_from_file("SF1_nomask.cimt")
        self.SF1 = imtio.import_from_file("SF1.cimt")
        self.VF1 = imtio.import_from_file("VF1.cimt")
        self.VF2 = imtio.import_from_file("VF2.cimt")
        self.Prof1 = imtio.import_from_file("Prof1.cimt")
        self.P1 = imtio.import_from_file("P1.cimt")
        self.TSF1 = imtio.import_from_file("TSF1.cimt")
        self.TSF1_nomask = imtio.import_from_file("TSF1_nomask.cimt")

    def test_imt_to_imts(self):
        imtio.imts_to_imt("VFs_imts_to_imt",
                          "TVF_imts_to_imt_a.cimt",
                          'TVF')
        TVF = TemporalVectorFields()
        TVF.add_field(self.VF1)
        TVF.add_field(self.VF2)
        TVF2 = imtio.import_from_file("TVF_imts_to_imt_a.cimt")
        assert TVF == TVF2

    def test_import_export_to_matlab(self):
        imtio.export_to_matlab(self.VF1, 'VF1.mat')
        VF1b = imtio.import_from_matlab('VF1.mat', 'VectorField',
                                        axe_x='axe_x',
                                        axe_y='axe_y',
                                        unit_x='unit_x',
                                        unit_y='unit_y',
                                        comp_x='comp_x',
                                        comp_y='comp_y',
                                        mask='mask',
                                        unit_values='unit_values')
        assert self.VF1 == VF1b
        #
        imtio.export_to_matlab(self.SF1, 'SF1.mat')
        SF1b = imtio.import_from_matlab('SF1.mat', 'ScalarField',
                                        axe_x='axe_x',
                                        axe_y='axe_y',
                                        unit_x='unit_x',
                                        unit_y='unit_y',
                                        values='values',
                                        mask='mask',
                                        unit_values='unit_values')
        assert self.SF1 == SF1b
        #
        imtio.export_to_matlab(self.Prof1, 'Prof1.mat')
        Prof1b = imtio.import_from_matlab('Prof1.mat', 'Profile',
                                          x='x',
                                          y='y',
                                          mask='mask',
                                          unit_x='unit_x',
                                          unit_y='unit_y')
        assert self.Prof1 == Prof1b
        #
        imtio.export_to_matlab(self.P1, 'P1.mat')
        P1b = imtio.import_from_matlab('P1.mat', 'Points',
                                       xy='xy',
                                       v='v',
                                       unit_x='unit_x',
                                       unit_y='unit_y',
                                       unit_v='unit_v')
        assert self.P1 == P1b

    def test_export_to_vtk(self):
        try:
            import pyvtk
        except ImportError:
            with pytest.raises(Exception):
                imtio.export_to_vtk(self.VF1, "VF1_export_to_vtk_a.cimt")
            with pytest.raises(Exception):
                imtio.export_to_vtk(self.SF1, "SF1_export_to_vtk_a.cimt")
            with pytest.raises(Exception):
                imtio.export_to_vtk(self.P1, "P1_export_to_vtk_a.cimt")
        else:
            imtio.export_to_vtk(self.VF1, "VF1_export_to_vtk_a.cimt")
            res_a2 = imtio.import_from_file("VF1_export_to_vtk_a.cimt")
            assert self.VF1 == res_a2
            #
            imtio.export_to_vtk(self.SF1, "SF1_export_to_vtk_a.cimt")
            res_a2 = imtio.import_from_file("SF1_export_to_vtk_a.cimt")
            assert self.SF1 == res_a2
            #
            imtio.export_to_vtk(self.P1, "P1_export_to_vtk_a.cimt")
            res_a2 = imtio.import_from_file("P1_export_to_vtk_a.cimt")
            assert self.P1 == res_a2

    def test_import_from_picture(self):
        imtio.export_to_picture(self.SF1_nomask, "SF1.png")
        res_a = imtio.import_from_picture("SF1.png")
        # imtio.export_to_file(res_a, "SF1_from_png.cimt")
        res_a2 = imtio.import_from_file("SF1_from_png.cimt")
        assert res_a == res_a2

    def test_import_from_pictures(self):
        imtio.export_to_pictures(self.TSF1_nomask, "TSF1")
        res_a = imtio.import_from_pictures("TSF1[0-9]+.png")
        # imtio.export_to_file(res_a, "TSF1_from_pngs.cimt")
        res_a2 = imtio.import_from_file("TSF1_from_pngs.cimt")
        assert res_a == res_a2

    def test_import_profile_from_ascii(self):
        imtio.export_to_ascii(self.Prof1, "Prof1.txt")
        res_a = imtio.import_profile_from_ascii("Prof1.txt", x_col=1,
                                                y_col=2,
                                                unit_x=self.Prof1.unit_x,
                                                unit_y=self.Prof1.unit_y)
        assert res_a == self.Prof1

    def test_import_sf_from_ascii(self):
        imtio.export_to_ascii(self.SF1, "SF1.txt")
        res_a = imtio.import_sf_from_ascii("SF1.txt",
                                           x_col=1,
                                           y_col=2,
                                           vx_col=3,
                                           unit_x=self.SF1.unit_x,
                                           unit_y=self.SF1.unit_y,
                                           unit_values=self.SF1.unit_values)
        assert res_a == self.SF1

    def test_import_vf_from_ascii(self):
        imtio.export_to_ascii(self.VF1, "VF1.txt")
        res_a = imtio.import_vf_from_ascii("VF1.txt",
                                           x_col=1,
                                           y_col=2,
                                           vx_col=3,
                                           vy_col=4,
                                           unit_x=self.VF1.unit_x,
                                           unit_y=self.VF1.unit_y,
                                           unit_values=self.VF1.unit_values)
        assert res_a == self.VF1

    def test_import_points_from_ascii(self):
        imtio.export_to_ascii(self.P1, "P1.txt")
        res_a = imtio.import_pts_from_ascii("P1.txt",
                                            x_col=1,
                                            y_col=2,
                                            v_col=3,
                                            unit_x=self.P1.unit_x,
                                            unit_y=self.P1.unit_y,
                                            unit_v=self.P1.unit_v)
        assert res_a == self.P1
        #
        imtio.export_to_ascii(self.P1, "P1_b.txt")
        res_a = imtio.import_pts_from_ascii("P1_b.txt",
                                            x_col=1,
                                            y_col=2,
                                            unit_x=self.P1.unit_x,
                                            unit_y=self.P1.unit_y)
        assert np.all(res_a.xy == self.P1.xy)
        assert res_a.unit_x == self.P1.unit_x
        assert res_a.unit_y == self.P1.unit_y



# TEMP
# pytest.main(['test_file_operation.py'])
# pytest.main(['-v', 'test_file_operation.py'])
# pytest.main(['--pdb', 'test_file_operation.py'])
# TEMP - End
